import typing as t
from pathlib import Path

import click
from click_option_group import optgroup, MutuallyExclusiveOptionGroup

from starwhale.consts import (
    SupportArch,
    PythonRunEnv,
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
)
from starwhale.base.uri import URI
from starwhale.base.type import URIType, RuntimeLockFileType
from starwhale.utils.cli import AliasedGroup
from starwhale.core.runtime.model import _SUPPORT_CUDA, _SUPPORT_CUDNN

from .view import get_term_view, RuntimeTermView
from .model import RuntimeInfoFilter


@click.group(
    "runtime",
    cls=AliasedGroup,
    help="Runtime management, quickstart/build/copy/activate...",
)
@click.pass_context
def runtime_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@click.group(
    "quickstart",
    cls=AliasedGroup,
    help="[Standalone]Quickstart your Starwhale Runtime",
)
def quickstart() -> None:
    pass


runtime_cmd.add_command(quickstart, aliases=["qs"])  # type: ignore


@quickstart.command("uri")
@click.argument("uri", required=True)
@click.argument("workdir", required=True)
@click.option("-f", "--force", is_flag=True, help="Force to quickstart")
@click.option("-n", "--name", default="", help="Runtime name")
@click.option(
    "-dr",
    "--disable-restore",
    is_flag=True,
    default=False,
    help="Create isolated python environment and restore the runtime",
)
def _quickstart_from_uri(
    uri: str, workdir: str, force: bool, name: str, disable_restore: bool
) -> None:
    """Quickstart from Starwhale Runtime URI

    Args:

        uri (str): Starwhale Dataset URI
        workdir (str): Runtime workdir
    """
    p_workdir = Path(workdir).absolute()
    name = name or p_workdir.name
    _uri = URI(uri, expected_type=URIType.RUNTIME)
    RuntimeTermView.quickstart_from_uri(
        workdir=p_workdir,
        name=name,
        uri=_uri,
        force=force,
        disable_restore=disable_restore,
    )


@quickstart.command("shell", help="Quickstart from interactive shell")
@click.argument("workdir", required=True)
@click.option("-f", "--force", is_flag=True, help="Force to quickstart")
@click.option(
    "-p",
    "--python-env",
    prompt="Choose your python env",
    type=click.Choice([PythonRunEnv.VENV, PythonRunEnv.CONDA]),
    default=PythonRunEnv.VENV,
    show_choices=True,
    show_default=True,
)
@click.option(
    "-n",
    "--name",
    default="",
    prompt="Please enter Starwhale Runtime name",
)
@click.option(
    "-dce",
    "--disable-create-env",
    prompt="Do you want to disable the isolated python environment creation automatically(NOT RECOMMENDED)?",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    default=False,
    help="Try entering the interactive shell at the end",
)
def _quickstart(
    workdir: str,
    force: bool,
    python_env: str,
    name: str,
    disable_create_env: bool,
    interactive: bool,
) -> None:
    """[Only Standalone]Quickstart Starwhale Runtime

    Args:

        workdir (Path): Runtime workdir
    """
    p_workdir = Path(workdir).absolute()
    name = name or p_workdir.name
    RuntimeTermView.quickstart_from_ishell(
        p_workdir, name, python_env, disable_create_env, force, interactive
    )


@runtime_cmd.command("build")
@optgroup.group(
    "\n  ** Acceptable build sources",
    cls=MutuallyExclusiveOptionGroup,
    help="The selector of the runtime build source, default is runtime.yaml source",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-c",
    "--conda",
    default="",
    help="from conda environment name",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-cp",
    "--conda-prefix",
    default="",
    help="from conda environment prefix path",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-v",
    "--venv",
    help="from virtualenv or python-venv environment prefix path",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-s",
    "--shell",
    is_flag=True,
    help="from current shell, venv or conda environment has been activated",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-y",
    "--yaml",
    help="from runtime yaml format file path.Default use runtime.yaml in the work directory(pwd)",
    default=DefaultYAMLName.RUNTIME,
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-d",
    "--docker",
    default="",
    help="from docker image",
)
@optgroup.group(
    "\n  ** Runtime YAML Source Configuration",
    help="The configurations only work for `--from-runtime-yaml` source",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-del",
    "--disable-env-lock",
    is_flag=True,
    help="Disable virtualenv/conda environment dependencies lock",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-nc",
    "--no-cache",
    is_flag=True,
    help="Invalid the cached(installed) packages in the isolate env when env-lock is enabled, \
    only for auto-generated environments",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-dad",
    "--download-all-deps",
    is_flag=True,
    help="Download all python dependencies into the runtime bundle file, the file size of swrt maybe very large.",
    hidden=True,
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-ie",
    "--include-editable",
    is_flag=True,
    help="Include editable packages",
    hidden=True,
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-ilw",
    "--include-local-wheel",
    is_flag=True,
    help="Include local wheel packages",
    hidden=True,
)
@optgroup.group(
    "\n  ** Conda/Venv/Shell Sources Configurations",
    help="The configurations only work for `--from-conda-name`, `--from-conda-prefix-path`, `--from-venv-prefix-path` and `--from-shell` sources",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "--cuda",
    type=click.Choice(_SUPPORT_CUDA + [""], case_sensitive=False),
    default="",
    help="cuda version, works for shell, conda name, conda prefix path and venv prefix path sources",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "--cudnn",
    default="",
    type=click.Choice(list(_SUPPORT_CUDNN.keys()) + [""], case_sensitive=False),
    help="cudnn version, works for shell, conda name, conda prefix path and venv prefix path sources",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "--arch",
    type=click.Choice(
        [SupportArch.AMD64, SupportArch.ARM64, SupportArch.NOARCH],
        case_sensitive=False,
    ),
    default=SupportArch.NOARCH,
    help="system architecture, works for shell, conda name, conda prefix path and venv prefix path sources",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-dad",
    "--download-all-deps",
    is_flag=True,
    help="Download all python dependencies into the runtime bundle file, the file size of swrt maybe very large.",
    hidden=True,
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-ie",
    "--include-editable",
    is_flag=True,
    help="Include editable packages",
    hidden=True,
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-ilw",
    "--include-local-wheel",
    is_flag=True,
    help="Include local wheel packages",
    hidden=True,
)
@optgroup.group("\n  ** Global Configurations")
@optgroup.option("-n", "--name", default="", help="runtime name")  # type: ignore[no-untyped-call]
@optgroup.option(  # type: ignore[no-untyped-call]
    "-p",
    "--project",
    default="",
    help="Project URI, default is the current selected project. The runtime package will store in the specified project.",
)
def _build(
    name: str,
    project: str,
    cuda: str,
    cudnn: str,
    arch: str,
    download_all_deps: bool,
    include_editable: bool,
    include_local_wheel: bool,
    disable_env_lock: bool,
    no_cache: bool,
    conda: str,
    conda_prefix: str,
    venv: str,
    shell: bool,
    yaml: str,
    docker: str,
) -> None:
    """Create and build a relocated, shareable, packaged runtime bundle(aka `swrt` file). Support python and native libs.
    Runtime build only works in the Standalone instance.

    Acceptable sources:

        \b
        - runtime.yaml file: The most flexible and customizable way.By the runtime.yaml file,
            you can specify the runtime name, python version, conda or venv environment, and the python dependency packages etc.
        - conda name: Lock the conda environment with conda name and generate the runtime bundle.
        - conda prefix path: Lock the conda environment with conda prefix path and generate the runtime bundle.
        - venv prefix path: Lock the virtualenv or python venv environment with venv prefix path and generate the runtime bundle.
        - shell: Lock the current shell environment and generate the runtime bundle. The current shell must be conda or venv.
        - docker image: Use the docker image as the runtime directly.

    Examples:

        \b
        - from runtime.yaml:
        swcli runtime build  # use the current directory as the workdir and use the default runtime.yaml file
        swcli runtime build -y example/pytorch/runtime.yaml # use example/pytorch/runtime.yaml as the runtime.yaml file
        swcli runtime build --yaml runtime.yaml # use runtime.yaml at the current directory as the runtime.yaml file

        \b
        - from conda name:
        swcli runtime build -c pytorch # lock pytorch conda environment and use `pytorch` as the runtime name
        swcli runtime build --conda pytorch --name pytorch-runtime # use `pytorch-runtime` as the runtime name
        swcli runtime build --conda pytorch --cuda 11.4 # specify the cuda version
        swcli runtime build --conda pytorch --arch noarch # specify the system architecture

        \b
        - from conda prefix path:
        swcli runtime build --conda-prefix /home/starwhale/anaconda3/envs/pytorch # get conda prefix path by `conda info --envs` command

        \b
        - from venv prefix path:
        swcli runtime build -v /home/starwhale/.virtualenvs/pytorch
        swcli runtime build --venv /home/starwhale/.local/share/virtualenvs/pytorch --arch amd64

        \b
        - from docker image:
        swcli runtime build --docker pytorch/pytorch:1.9.0-cuda11.1-cudnn8-runtime  # use the docker image as the runtime directly

        \b
        - from shell:
        swcli runtime build -s --cuda 11.4 --cudnn 8 # specify the cuda and cudnn version
        swcli runtime build --shell --name pytorch-runtime # lock the current shell environment and use `pytorch-runtime` as the runtime name
    """
    if docker:
        RuntimeTermView.build_from_docker_image(
            image=docker, runtime_name=name, project=project
        )
    elif conda or conda_prefix or venv or shell:
        # TODO: support auto mode for cuda, cudnn and arch
        RuntimeTermView.build_from_python_env(
            runtime_name=name,
            conda_name=conda,
            conda_prefix=conda_prefix,
            venv_prefix=venv,
            project=project,
            cuda=cuda,
            cudnn=cudnn,
            arch=arch,
            download_all_deps=download_all_deps,
            include_editable=include_editable,
            include_local_wheel=include_local_wheel,
        )
    else:
        RuntimeTermView.build_from_runtime_yaml(
            workdir=Path.cwd(),
            yaml_path=Path(yaml),
            project=project,
            runtime_name=name,
            download_all_deps=download_all_deps,
            include_editable=include_editable,
            include_local_wheel=include_local_wheel,
            no_cache=no_cache,
            disable_env_lock=disable_env_lock,
        )


@runtime_cmd.command("remove", aliases=["rm"])
@click.argument("runtime")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force to remove runtime, the removed runtime cannot recover",
)
def _remove(runtime: str, force: bool) -> None:
    """
    Remove runtime

    You can run `swcli runtime recover` to recover the removed runtimes.

    RUNTIME: argument use the `Runtime URI` format, so you can remove the whole runtime or a specified-version runtime.
    """
    click.confirm("continue to remove?", abort=True)
    RuntimeTermView(runtime).remove(force)


@runtime_cmd.command("recover")
@click.argument("runtime")
@click.option("-f", "--force", is_flag=True, help="Force to recover runtime")
def _recover(runtime: str, force: bool) -> None:
    """
    Recover runtime

    RUNTIME: argument use the `Runtime URI` format, so you can recover the whole runtime or a specified-version runtime.
    """
    RuntimeTermView(runtime).recover(force)


@runtime_cmd.command("info")
@click.argument("runtime")
@click.option(
    "-of",
    "--output-filter",
    type=click.Choice([f.value for f in RuntimeInfoFilter], case_sensitive=False),
    default=RuntimeInfoFilter.basic.value,
    show_default=True,
    help="Filter the output content. Only standalone instance supports this option.",
)
@click.pass_obj
def _info(
    view: t.Type[RuntimeTermView],
    runtime: str,
    output_filter: str,
) -> None:
    """Show runtime details

    RUNTIME: argument use the `Runtime URI` format. Version is optional for the Runtime URI.
    If the version is not specified, the latest version will be used.

    Example:

        \b
          swcli runtime info pytorch # show basic info from the latest version of runtime
          swcli runtime info pytorch/version/v0  # show basic info
          swcli runtime info pytorch/version/v0 --output-filter basic  # show basic info
          swcli runtime info pytorch/version/v1 -of runtime_yaml  # show runtime.yaml content
          swcli runtime info pytorch/version/v1 -of lock # show auto lock file content
          swcli runtime info pytorch/version/v1 -of manifest # show _manifest.yaml content
          swcli runtime info pytorch/version/v1 -of all # show all info of the runtime
    """
    uri = URI(runtime, expected_type=URIType.RUNTIME)
    if not uri.object.version:
        uri.object.version = "latest"

    view(uri).info(RuntimeInfoFilter(output_filter))


@runtime_cmd.command("history", help="Show runtime history")
@click.argument("runtime", required=True)
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _history(view: t.Type[RuntimeTermView], runtime: str, fullname: bool) -> None:
    view(runtime).history(fullname)


# hide runtime restore command for the users in the command help output.
@runtime_cmd.command("restore", hidden=True)
@click.argument("target")
def _restore(target: str) -> None:
    """
    [Only Standalone]Prepare dirs, restore python environment with virtualenv or conda and show activate command.

    TARGET: runtime uri or runtime workdir path, in Starwhale Agent Docker Environment, only support workdir path.
    """
    RuntimeTermView.restore(target)


@runtime_cmd.command("list", aliases=["ls"])
@click.option(
    "-p",
    "--project",
    default="",
    help="Project URI, the default is the current selected project.",
)
@click.option("-f", "--fullname", is_flag=True, help="Show fullname of runtime version")
@click.option("-sr", "--show-removed", is_flag=True, help="Show removed runtime")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for tasks list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for tasks list"
)
@click.option(
    "filters",
    "-fl",
    "--filter",
    multiple=True,
    help="Filter output based on conditions provided.",
)
@click.pass_obj
def _list(
    view: t.Type[RuntimeTermView],
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
    filters: list,
) -> None:
    """
    List Runtime of the specified project

    The filtering flag (-fl or --filter) format is a key=value pair or a flag.
    If there is more than one filter, then pass multiple flags.\n
    (e.g. --filter name=mnist --filter latest)

    \b
    The currently supported filters are:
      name\tTEXT\tThe prefix of the runtime name
      owner\tTEXT\tThe name or id of the runtime owner
      latest\tFLAG\t[Cloud] Only show the latest version
            \t \t[Standalone] Only show the version with "latest" tag
    """
    view.list(project, fullname, show_removed, page, size, filters)


@runtime_cmd.command(
    "extract", help="[Only Standalone]Extract local runtime tar file into workdir"
)
@click.argument("runtime")
@click.option("-f", "--force", is_flag=True, help="Force to extract runtime")
@click.option(
    "--target-dir",
    default="",
    help="Extract target dir.if omitted, sw will use starwhale default workdir",
)
def _extract(runtime: str, force: bool, target_dir: str) -> None:
    RuntimeTermView(runtime).extract(force, target_dir)


@runtime_cmd.command("copy", aliases=["cp"])
@click.argument("src")
@click.argument("dest")
@click.option("-f", "--force", is_flag=True, help="Force to copy")
@click.option("-dlp", "--dest-local-project", help="dest local project uri")
def _copy(src: str, dest: str, force: bool, dest_local_project: str) -> None:
    """
    Copy Runtime between Standalone Instance and Cloud Instance

    SRC: runtime uri with version

    DEST: project uri or runtime uri with name.

    Example:

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud runtime to local project(myproject) with a new runtime name 'mnist-local'
            swcli runtime cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq local/project/myproject/mnist-local

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud runtime to local default project(self) with the cloud instance runtime name 'mnist-cloud'
            swcli runtime cp cloud://pre-k8s/project/runtime/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq .

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud runtime to local project(myproject) with the cloud instance runtime name 'mnist-cloud'
            swcli runtime cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq . -dlp myproject

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud runtime to local default project(self) with a runtime name 'mnist-local'
            swcli runtime cp cloud://pre-k8s/project/runtime/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq mnist-local

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud runtime to local project(myproject) with a runtime name 'mnist-local'
            swcli runtime cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq mnist-local -dlp myproject

        \b
        - copy standalone instance(local) default project(self)'s mnist-local runtime to cloud instance(pre-k8s) mnist project with a new runtime name 'mnist-cloud'
            swcli runtime cp mnist-local/version/latest cloud://pre-k8s/project/mnist/mnist-cloud

        \b
        - copy standalone instance(local) default project(self)'s mnist-local runtime to cloud instance(pre-k8s) mnist project with standalone instance runtime name 'mnist-local'
            swcli runtime cp mnist-local/version/latest cloud://pre-k8s/project/mnist

        \b
        - copy standalone instance(local) default project(self)'s mnist-local runtime to cloud instance(pre-k8s) mnist project without 'cloud://' prefix
            swcli runtime cp mnist-local/version/latest pre-k8s/project/mnist

        \b
        - copy standalone instance(local) project(myproject)'s mnist-local runtime to cloud instance(pre-k8s) mnist project with standalone instance runtime name 'mnist-local'
            swcli runtime cp local/project/myproject/runtime/mnist-local/version/latest cloud://pre-k8s/project/mnist
    """
    RuntimeTermView.copy(src, dest, force, dest_local_project)


@runtime_cmd.command("tag", help="Runtime Tag Management, add or remove")
@click.argument("runtime")
@click.argument("tags", nargs=-1)
@click.option("-r", "--remove", is_flag=True, help="Remove tags")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Ignore tag name errors like name duplication, name absence",
)
def _tag(runtime: str, tags: t.List[str], remove: bool, quiet: bool) -> None:
    RuntimeTermView(runtime).tag(tags, remove, quiet)


@runtime_cmd.command(
    "activate",
    aliases=["actv"],
    help="",
)
@click.argument("uri")
@click.option(
    "-f",
    "--force-restore",
    help="Force to restore runtime into the related snapshot workdir even the runtime has been restored",
)
def _activate(uri: str, force_restore: bool) -> None:
    """
    [Only Standalone]Activate python runtime environment for development

    When the runtime has not been restored, activate command will restore runtime automatically.

    URI: Runtime uri in the standalone instance
    """
    _uri = URI(uri, expected_type=URIType.RUNTIME)
    RuntimeTermView.activate(_uri, force_restore)


@runtime_cmd.command("lock")
@click.argument("target_dir", default=".")
@click.option(
    "-f",
    "--yaml-path",
    default=DefaultYAMLName.RUNTIME,
    help=f"Runtime YAML file path, default is {DefaultYAMLName.RUNTIME} at the current working directory",
)
@optgroup.group(  # type: ignore
    "Python environment selectors",
    cls=MutuallyExclusiveOptionGroup,
    help="The selector of the python environment, default is the starwhale auto create env prefix path",
)
@optgroup.option("-n", "--env-name", default="", help="conda name")  # type: ignore
@optgroup.option(  # type: ignore
    "-p", "--env-prefix-path", default="", help="conda or virtualenv prefix path"
)
@optgroup.option(  # type: ignore
    "-s", "--env-use-shell", is_flag=True, default=False, help="use current shell"
)
@click.option(
    "-so", "--stdout", is_flag=True, help="Output lock file content to the stdout"
)
@click.option(
    "-ie",
    "--include-editable",
    is_flag=True,
    help="Include editable packages, only for venv mode",
)
@click.option(
    "-ilw",
    "--include-local-wheel",
    is_flag=True,
    help="Include local wheel packages, only for venv mode",
)
@click.option(
    "-epo",
    "--emit-pip-options",
    is_flag=True,
    help=f"Emit pip config options when the command dumps {RuntimeLockFileType.VENV}",
)
@click.option(
    "-nc",
    "--no-cache",
    is_flag=True,
    help="Invalid the cached(installed) packages in the isolate env when env-lock is enabled, \
    only for auto-generated environments",
)
def _lock(
    target_dir: str,
    yaml_path: str,
    env_name: str,
    env_prefix_path: str,
    env_use_shell: bool,
    stdout: bool,
    include_editable: bool,
    include_local_wheel: bool,
    emit_pip_options: bool,
    no_cache: bool,
) -> None:
    """
    [Only Standalone]Lock Python venv or conda environment

    TARGET_DIR: the lock files will store in the `target_dir` , default is "."
    """

    RuntimeTermView.lock(
        target_dir,
        Path(yaml_path),
        env_name,
        env_prefix_path,
        no_cache,
        stdout,
        include_editable,
        include_local_wheel,
        emit_pip_options,
        env_use_shell,
    )


@runtime_cmd.command("dockerize")
@click.argument("uri", required=True)
@click.option("-t", "--tag", multiple=True, help="Image tag")
@click.option("--push", is_flag=True, help="Push image to the registry")
@click.option(
    "--platform",
    multiple=True,
    default=[SupportArch.AMD64],
    type=click.Choice([SupportArch.AMD64, SupportArch.ARM64]),
    help="Target platform for docker build",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="Only render Dockerfile and build command",
)
@click.option(
    "--use-starwhale-builder",
    is_flag=True,
    help="Starwhale will create buildx builder for multi-arch",
)
@click.option(
    "--reset-qemu-static",
    is_flag=True,
    help="Reset qemu static, then fix multiarch build issue",
)
def _dockerize(
    uri: str,
    tag: t.Tuple[str],
    push: bool,
    platform: t.Tuple[str],
    dry_run: bool,
    use_starwhale_builder: bool,
    reset_qemu_static: bool,
) -> None:
    """[Only Standalone]Starwhale runtime dockerize, only for standalone instance

    URI (str): Starwhale Runtime URI in the standalone instance
    """
    RuntimeTermView(uri).dockerize(
        list(tag),
        push,
        list(platform),
        dry_run,
        use_starwhale_builder,
        reset_qemu_static,
    )



from __future__ import annotations

import io
import os
import queue
import struct
import typing as t
import tempfile
import threading
from types import TracebackType
from pathlib import Path
from binascii import crc32
from collections import defaultdict

from loguru import logger

from starwhale.consts import STANDALONE_INSTANCE
from starwhale.base.uri import URI
from starwhale.utils.fs import (
    empty_dir,
    ensure_dir,
    blake2b_file,
    BLAKE2B_SIGNATURE_ALGO,
)
from starwhale.base.type import URIType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.retry import http_retry
from starwhale.core.dataset.type import (
    Link,
    BaseArtifact,
    D_ALIGNMENT_SIZE,
    D_FILE_VOLUME_SIZE,
)
from starwhale.core.dataset.store import DatasetStorage
from starwhale.core.dataset.tabular import TabularDataset, TabularDatasetRow
from starwhale.api._impl.dataset.loader import DataRow


class RotatedBinWriter:
    """
    format:
        header_magic    uint32  I
        crc             uint32  I
        _reserved       uint64  Q
        size            uint32  I
        padding_size    uint32  I
        header_version  uint32  I
        data_magic      uint32  I --> above 32 bytes
        data bytes...
        padding bytes...        --> default 4K padding
    """

    _header_magic = struct.unpack(">I", b"SWDS")[0]
    _data_magic = struct.unpack(">I", b"SDWS")[0]
    _header_struct = struct.Struct(">IIQIIII")
    _header_size = _header_struct.size
    _header_version = 0

    class _BinSection(t.NamedTuple):
        offset: int
        size: int
        raw_data_offset: int
        raw_data_size: int

    def __init__(
        self,
        workdir: Path,
        rotated_bin_notify_queue: t.Optional[queue.Queue[t.Optional[Path]]] = None,
        alignment_bytes_size: int = D_ALIGNMENT_SIZE,
        volume_bytes_size: int = D_FILE_VOLUME_SIZE,
    ) -> None:
        self.workdir = workdir

        if alignment_bytes_size <= 0:
            raise ValueError(
                f"alignment_bytes_size must be greater than zero: {alignment_bytes_size}"
            )

        self.alignment_bytes_size = alignment_bytes_size

        if volume_bytes_size < 0:
            raise ValueError(
                f"volume_bytes_size must be greater than zero: {volume_bytes_size}"
            )
        self.volume_bytes_size = volume_bytes_size

        self._rotated_bin_notify_queue = rotated_bin_notify_queue
        self._wrote_size = 0

        ensure_dir(self.workdir)
        _, bin_writer_path = tempfile.mkstemp(
            prefix="bin-writer-", dir=str(self.workdir.absolute())
        )
        self._current_writer_path = Path(bin_writer_path)
        self._current_writer = self._current_writer_path.open("wb")
        self._rotated_bin_paths: t.List[Path] = []

    def __enter__(self) -> RotatedBinWriter:
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:  # pragma: no cover
            logger.warning(f"type:{type}, exception:{value}, traceback:{trace}")

        self.close()

    @property
    def working_path(self) -> Path:
        return self._current_writer_path

    @property
    def rotated_paths(self) -> t.List[Path]:
        return self._rotated_bin_paths

    def close(self) -> None:
        self._rotate()

        empty = self._current_writer.tell() == 0
        self._current_writer.close()
        if empty and self._current_writer_path.exists():
            # last file is empty
            os.unlink(self._current_writer_path)

    def _rotate(self) -> None:
        if self._wrote_size == 0:
            return
        self._current_writer.close()
        self._wrote_size = 0

        self._rotated_bin_paths.append(self._current_writer_path)
        if self._rotated_bin_notify_queue is not None:
            self._rotated_bin_notify_queue.put(self._current_writer_path)

        _, bin_writer_path = tempfile.mkstemp(
            prefix="bin-writer-", dir=str(self.workdir.absolute())
        )
        self._current_writer_path = Path(bin_writer_path)
        self._current_writer = self._current_writer_path.open("wb")

    def write(self, data: bytes) -> t.Tuple[Path, RotatedBinWriter._BinSection]:
        _cls = RotatedBinWriter
        size = len(data)
        crc = crc32(data)  # TODO: crc is right?
        start = self._current_writer.tell()
        padding_size = self._get_padding_size(size)

        _header = _cls._header_struct.pack(
            _cls._header_magic,
            crc,
            0,
            size,
            padding_size,
            _cls._header_version,
            _cls._data_magic,
        )
        _padding = b"\0" * padding_size
        self._current_writer.write(_header + data + _padding)
        _bin_path = self._current_writer_path
        _bin_section = _cls._BinSection(
            offset=start,
            size=_cls._header_size + size + padding_size,
            raw_data_offset=start + _cls._header_size,
            raw_data_size=size,
        )
        self._wrote_size += _bin_section.size

        if self._wrote_size > self.volume_bytes_size:
            self._rotate()

        return _bin_path, _bin_section

    def _get_padding_size(self, size: int) -> int:
        remain = (size + RotatedBinWriter._header_size) % self.alignment_bytes_size
        return 0 if remain == 0 else (self.alignment_bytes_size - remain)


class MappingDatasetBuilder:
    _STASH_URI = "_starwhale_stash_uri"

    class _SignedBinMeta(t.NamedTuple):
        name: str
        algo: str
        size: int

        def __str__(self) -> str:
            return f"{self.name}:{self.algo}:{self.size}"

    def __init__(
        self,
        workdir: t.Union[Path, str],
        dataset_name: str,
        project_name: str,
        instance_name: str = STANDALONE_INSTANCE,
        blob_alignment_bytes_size: int = D_ALIGNMENT_SIZE,
        blob_volume_bytes_size: int = D_FILE_VOLUME_SIZE,
    ) -> None:
        self.workdir = Path(workdir)
        self.dataset_name = dataset_name
        self.project_name = project_name
        self.instance_name = instance_name

        self._in_standalone = instance_name == STANDALONE_INSTANCE

        self._blob_alignment_bytes_size = blob_alignment_bytes_size
        self._blob_volume_bytes_size = blob_volume_bytes_size
        self._tabular_dataset = TabularDataset(
            name=dataset_name,
            project=project_name,
            instance_name=instance_name,
        )
        self._rows_put_queue: queue.Queue[
            t.Optional[t.Union[DataRow, Exception]]
        ] = queue.Queue()
        self._rows_put_thread = threading.Thread(
            target=self._rows_put_worker, daemon=True, name="RowPutThread"
        )
        self._rows_put_thread.start()
        self._rows_put_exception: t.Optional[Exception] = None

        # abs = artifacts bin sync
        self._abs_queue: queue.Queue[t.Optional[Path]] = queue.Queue()
        self._abs_exception: t.Optional[Exception] = None
        self._abs_thread = threading.Thread(
            target=self._abs_worker, daemon=True, name="ArtifactsBinSyncThread"
        )
        self._abs_thread.start()

        self._artifact_bin_tmpdir = self.workdir / "artifact_bin_tmp"
        self._artifact_bin_writer = RotatedBinWriter(
            workdir=self._artifact_bin_tmpdir,
            rotated_bin_notify_queue=self._abs_queue,
            alignment_bytes_size=self._blob_alignment_bytes_size,
            volume_bytes_size=self._blob_volume_bytes_size,
        )

        self._stash_uri_rows_map: t.Dict[
            Path, t.List[t.Tuple[BaseArtifact, TabularDatasetRow]]
        ] = defaultdict(list)
        self._signed_bins_meta: t.List[MappingDatasetBuilder._SignedBinMeta] = []
        self._last_flush_revision = ""

    def __enter__(self) -> MappingDatasetBuilder:
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:  # pragma: no cover
            logger.warning(f"type:{type}, exception:{value}, traceback:{trace}")

        self.close()

    def put(self, row: DataRow) -> None:
        if not row or not isinstance(row, DataRow):
            raise ValueError(f"row argument must be DataRow type: {row}")

        self._raise_thread_daemons_exception()
        self._rows_put_queue.put(row)

    def _raise_thread_daemons_exception(self) -> None:
        if self._abs_exception:
            _e = self._abs_exception
            self._abs_exception = None
            raise threading.ThreadError(f"ArtifactsBinSyncThread raise exception: {_e}")

        if self._rows_put_exception:
            _e = self._rows_put_exception
            self._rows_put_exception = None
            raise threading.ThreadError(f"RowPutThread raise exception: {_e}")

    def _handle_row_put(self, row: DataRow) -> None:
        td_row = TabularDatasetRow(id=row.index, features=row.features)

        for artifact in td_row.artifacts:
            # TODO: refactor BaseArtifact Type, parse link by fp, such as: fp="s3://xx/yy/zz", fp="http://xx/yy/zz"
            if not artifact.link:
                if isinstance(artifact.fp, (str, Path)):
                    content = Path(artifact.fp).read_bytes()
                elif isinstance(artifact.fp, (bytes, io.IOBase)):
                    content = artifact.to_bytes()
                else:
                    raise TypeError(
                        f"no support fp type for bin writer:{type(artifact.fp)}, {artifact.fp}"
                    )

                _path, _meta = self._artifact_bin_writer.write(content)
                artifact.link = Link(
                    uri=self._STASH_URI,  # When bin writer is rotated, we can get the signatured uri
                    offset=_meta.raw_data_offset,
                    size=_meta.raw_data_size,
                    bin_offset=_meta.offset,
                    bin_size=_meta.size,
                )
                self._stash_uri_rows_map[_path].append((artifact, td_row))

            # TODO: find a graceful cleanup method
            # forbid to write cache and fp contents into datastore table
            artifact.clear_bytes()

        self._tabular_dataset.put(td_row)

    def _handle_bin_sync(self, bin_path: Path) -> None:
        size = bin_path.stat().st_size
        if self._in_standalone:
            uri, _ = DatasetStorage.save_data_file(bin_path, remove_src=True)
        else:
            sign_name = blake2b_file(bin_path)
            crm = CloudRequestMixed()
            instance_uri = URI(self.instance_name, expected_type=URIType.INSTANCE)

            @http_retry
            def _upload() -> str:
                r = crm.do_multipart_upload_file(
                    url_path=f"/project/{self.project_name}/dataset/{self.dataset_name}/hashedBlob/{sign_name}",
                    file_path=bin_path,
                    instance_uri=instance_uri,
                )
                return r.json()["data"]  # type: ignore

            uri = _upload()

        self._signed_bins_meta.append(
            MappingDatasetBuilder._SignedBinMeta(
                name=uri,
                algo=BLAKE2B_SIGNATURE_ALGO,
                size=size,
            )
        )

        for artifact, td_row in self._stash_uri_rows_map.get(bin_path, []):
            artifact.link.uri = uri  # type: ignore
            self._tabular_dataset.put(td_row)

        if bin_path in self._stash_uri_rows_map:
            del self._stash_uri_rows_map[bin_path]

    def _rows_put_worker(self) -> None:
        try:
            while True:
                row = self._rows_put_queue.get(block=True, timeout=None)
                if row is None:
                    break
                elif isinstance(
                    row, Exception
                ):  # receive exception from ArtifactsBinSyncThread
                    raise row
                else:
                    try:
                        self._handle_row_put(row)
                    finally:
                        self._rows_put_queue.task_done()
        except Exception as e:
            self._rows_put_exception = e
            raise
        finally:
            self._artifact_bin_writer.close()
            self._abs_queue.put(None)

    def _abs_worker(self) -> None:
        try:
            while True:
                bin_path = self._abs_queue.get(block=True, timeout=None)
                if bin_path is None:
                    break
                else:
                    try:
                        self._handle_bin_sync(bin_path)
                    finally:
                        self._abs_queue.task_done()
        except Exception as e:
            self._abs_exception = e
            self._rows_put_queue.put(e)
            raise

    def delete(self, key: t.Union[str, int]) -> None:
        self._tabular_dataset.delete(key)

    def flush(self, artifacts_flush: bool = False) -> str:
        self._rows_put_queue.join()

        if artifacts_flush:
            self._artifact_bin_writer._rotate()
            self._abs_queue.join()

        self._last_flush_revision, _ = self._tabular_dataset.flush()
        return self._last_flush_revision

    def _threads_join(self) -> None:
        self._abs_thread.join()
        self._rows_put_thread.join()

    def close(self) -> None:
        self._rows_put_queue.put(None)
        self._threads_join()
        self._tabular_dataset.close()

        empty_dir(self._artifact_bin_tmpdir)
        self._raise_thread_daemons_exception()

    @property
    def signature_bins_meta(self) -> t.List[MappingDatasetBuilder._SignedBinMeta]:
        return self._signed_bins_meta

    def calculate_rows_cnt(self) -> int:
        # TODO: tune performance by datastore
        return len(
            [
                row
                for row in self._tabular_dataset.scan(
                    revision=self._last_flush_revision
                )
            ]
        )



import re
from enum import Enum
from glob import glob
from typing import Any, Dict, Optional
from pathlib import Path
from dataclasses import dataclass

import requests

from starwhale.consts import SW_API_VERSION
from starwhale.base.uri import URI
from starwhale.utils.config import load_swcli_config
from starwhale.base.uricomponents.project import Project
from starwhale.base.uricomponents.instance import Instance
from starwhale.base.uricomponents.exceptions import (
    NoMatchException,
    UriTooShortException,
)

url_regex = re.compile(
    r"(?P<scheme>https*)://"
    r"(?P<host>.*)/projects/"
    r"(?P<project>.+)/"
    r"(?P<rc_type>models|datasets|runtimes|evaluations)"
    r"(/(?P<rc_id>.+)/versions)?"  # optional
    r"(/(?P<rc_version>.+)/.*)?",  # optional
    re.UNICODE,
)


class ResourceType(Enum):
    runtime = "runtime"
    model = "model"
    dataset = "dataset"
    evaluation = "evaluation"


@dataclass
class Resource:
    """
    Resource holds the DataSet, Model, Runtime, Eval etc.
    """

    typ: ResourceType
    project: Project
    name: Optional[str] = None
    version: Optional[str] = None

    def __init__(
        self,
        uri: str,
        typ: Optional[ResourceType] = None,
        project: Optional[Project] = None,
    ) -> None:
        """
        :param uri: resource name or uri or version
            accept patterns:
                * resource name without type, e.g. mnist
                * version e.g. gntdqm3fgeydqnrtmftdgyjzpj4hamy
                * full uri without version e.g. local/project/self/dataset/mnist
                * full uri e.g. local/project/self/dataset/mnist/version/latest
                * resource name and version e.g. mnist/version/latest
            typically unacceptable cases:
                * resource type and name under project e.g. self/dataset/mnist
                * resource type, name and version under project e.g. self/dataset/mnist/version/latest
        :param typ: resource type: (dataset, model, runtime etc)
        :param project: project which the resource belongs to (optional)
        :return: Resource instance
        """
        self._remote_info: Dict[str, Any] = {}

        # check if it is url from console
        m = url_regex.match(uri)
        if m is not None:
            info: Dict[str, str] = m.groupdict()
            ins = Instance(f'{info["scheme"]}://{info["host"]}')
            self.project = Project(name=info["project"], instance=ins)
            self.typ = ResourceType(info["rc_type"][:-1])  # remove the last 's'
            self.name = info.get("rc_id")
            self.version = info.get("rc_version")
            self.refine_remote_rc_info()
            return

        if project:
            self.project = project
        else:
            try:
                self.project = Project.parse_from_full_uri(
                    uri, ignore_rc_type=typ is not None
                )
                uri = self.project.path
            except UriTooShortException:
                self.project = Project()

        if "://" in uri:
            uri = uri.split("://", 1)[-1]

        if "//" in uri:
            raise Exception(f"wrong uri({uri}) format with '//'")

        parts = len(uri.split("/"))
        # 3 == len('mnist/version/latest'.split(/))
        # means that there is no type info in the uri
        if typ and parts <= 3:
            self._parse_with_type(typ, uri)
        else:
            # uri is long enough to parse type
            self._parse_without_type(uri)
        if (typ is not None) and self.typ != typ:
            raise Exception(f"type mismatch: {self.typ} vs {typ}")

        if not self.project.instance.is_local:
            self.refine_remote_rc_info()

    def _parse_with_type(self, typ: ResourceType, uri: str) -> None:
        """
        :param typ:
        :param uri:
            i.e.
            * mnist
            * mnist/latest
            * version/latest
            * mnist/version/latest
        """
        self.typ = typ
        parts = uri.split("/")
        if len(parts) > 3:
            raise Exception(f"invalid uri {uri} when parse with type {typ}")

        if len(parts) == 1:
            try:
                # try version first
                self._parse_by_version(parts[0])
            except NoMatchException:
                self.name = parts[0]
        elif len(parts) == 2:
            if parts[0] != "version":
                # name/version
                self.name = parts[0]
            self.version = parts[-1]
        else:
            if parts[1] != "version":
                raise Exception(f"invalid uri without version field: {uri}")
            self.name = parts[0]
            self.version = parts[-1]

    def _parse_without_type(self, uri: str) -> None:
        p = uri.split("/", 1)
        if len(p) == 1:
            # version only (we do not support name without type parsing)
            # there is no scenario for 'name without type' for now
            return self._parse_by_version(uri)
        typ = ResourceType[p[0]]
        return self._parse_with_type(typ, p[1])

    def _parse_by_version(self, ver: str) -> None:
        if self.instance.is_local:
            root = Path(load_swcli_config()["storage"]["root"]) / self.project.name
            # storage-root/project/type/name/prefix/full-version
            p = f"{root.absolute()}/*/*/*/{ver}*"
            m = glob(p)
            if len(m) == 1:
                _, typ, self.name, _, self.version = m[0].rsplit("/", 4)
                self.typ = ResourceType[typ]
            else:
                raise NoMatchException(ver, list(m))
        else:
            # TODO use api to check if it is a name or version
            # assume is is name for now
            self.name = ver

    def refine_remote_rc_info(self) -> None:
        if self.project.instance.is_local:
            raise Exception("only used for remote resources")
        if not self.name or not self.version:
            return
        if not self.name.isnumeric() and not self.version.isnumeric():
            # both are not numeric, assume it is already refined
            return

        base_path = f"{self.instance.url}/api/{SW_API_VERSION}/project/{self.project.name}/{self.typ.value}/{self.name}"
        headers = {"Authorization": self.instance.token}
        resp = requests.get(
            base_path, timeout=60, params={"versionUrl": self.version}, headers=headers
        )
        resp.raise_for_status()
        self._remote_info = resp.json().get("data", {})
        self.name = self._remote_info.get("name", self.name)
        self.version = self._remote_info.get("versionName", self.version)

    def info(self) -> Dict[str, Any]:
        # TODO: support local resource
        return self._remote_info or {}

    @property
    def instance(self) -> Instance:
        return self.project.instance

    def to_uri(self) -> URI:
        return URI.capsulate_uri(
            instance=str(self.instance),
            project=self.project.name,
            obj_type=self.typ.name,
            obj_name=self.name or "",
            obj_ver=self.version or "latest",
        )

    def __str__(self) -> str:
        return f"project: {self.project}, name: {self.name}, version: {self.version}"

    __repr__ = __str__



import os
import sys
import typing as t
import getpass as gt
import subprocess
from pwd import getpwnam
from pathlib import Path

from starwhale.utils import config, console
from starwhale.consts import SupportArch, CNTR_DEFAULT_PIP_CACHE_DIR
from starwhale.utils.error import NoSupportError, MissingFieldError
from starwhale.utils.process import check_call

BUILDX_ARCH_MAP = {SupportArch.AMD64: "linux/amd64", SupportArch.ARM64: "linux/arm64"}
BUILDER_NAME = "starwhale-multiarch-runtime-builder"

# make Docker-ce 19.03 happy to use buildx
_BUILDX_CMD_ENV = {"DOCKER_CLI_EXPERIMENTAL": "enabled"}


def reset_qemu_static() -> None:
    # ref: https://github.com/multiarch/qemu-user-static
    check_call(
        [
            "docker",
            "run",
            "--rm",
            "--privileged",
            "multiarch/qemu-user-static",
            "--reset",
            "-p",
            "-yes",
        ],
        log=console.print,
    )


def check_builder_exists() -> bool:
    cmd = [
        "docker",
        "buildx",
        "inspect",
        "--builder",
        BUILDER_NAME,
    ]
    try:
        check_call(cmd, env=_BUILDX_CMD_ENV)
    except subprocess.CalledProcessError as e:
        if "no builder" in e.output:
            return False
        raise

    return True


def create_builder(dry_run: bool) -> None:
    # ref: https://docs.docker.com/build/buildx/drivers/ for multi-arch build
    if not dry_run and check_builder_exists():
        console.print(f"{BUILDER_NAME} already exists, skip create")
        return

    cmd = [
        "docker",
        "buildx",
        "create",
        "--bootstrap",
        "--name",
        BUILDER_NAME,
        "--driver",
        "docker-container",
    ]

    _http_proxy = os.environ.get("HTTP_PROXY") or os.environ.get("http_proxy")
    _https_proxy = os.environ.get("HTTPS_PROXY") or os.environ.get("https_proxy")
    if _http_proxy:
        cmd += [
            "--driver-opt",
            f"env.http_proxy={_http_proxy}",
            "--driver-opt",
            f"env.HTTP_PROXY={_http_proxy}",
        ]

    if _https_proxy:
        cmd += [
            "--driver-opt",
            f"env.https_proxy={_https_proxy}",
            "--driver-opt",
            f"env.HTTPS_PROXY={_https_proxy}",
        ]

    cmd += ["--driver-opt", "network=host"]

    if dry_run:
        console.print(":tiger_face: buildx create command:")
        console.print(f"\t [bold red]{' '.join(cmd)}[/]")
    else:
        console.print(":tiger_face: start to create buildx instance...")
        check_call(cmd, log=console.print, env=_BUILDX_CMD_ENV)


def buildx(
    dockerfile_path: Path,
    workdir: Path,
    tags: t.List[str],
    platforms: t.List[str],
    push: bool = False,
    dry_run: bool = False,
    use_starwhale_builder: bool = False,
) -> None:
    if not tags:
        raise MissingFieldError(
            "buildx must contain at least one or more tag parameter"
        )

    cmd = [
        "docker",
        "buildx",
        "build",
        "--network",
        "host",
        "--pull",
        "--progress",
        "plain",
    ]

    if use_starwhale_builder:
        create_builder(dry_run)
        cmd += ["--builder", BUILDER_NAME]

    for _p in platforms:
        # TODO: add multiarch/qemu-user-static reset for non-amd64 arch
        if _p not in BUILDX_ARCH_MAP:
            raise NoSupportError(f"Starwhale base image {_p} arch")
        cmd += ["--platform", BUILDX_ARCH_MAP[_p]]

    for _t in tags:
        _t = _t.strip()
        if not _t:
            continue
        cmd += ["--tag", _t]

    if push:
        cmd += ["--push"]
    else:
        cmd += ["--load"]

    cmd += ["--file", str(dockerfile_path.absolute()), str(workdir.absolute())]

    if dry_run:
        console.print(":panda_face: buildx build command:")
        console.print(f"\t [bold red]{' '.join(cmd)}[/]")
    else:
        console.print(":panda_face: start to build image with buildx...")
        check_call(cmd, log=console.print, env=_BUILDX_CMD_ENV)


def gen_swcli_docker_cmd(
    image: str,
    env_vars: t.Dict[str, str] = {},
    mnt_paths: t.List[str] = [],
    name: str = "",
) -> str:

    if not image:
        raise ValueError("image should have value")
    pwd = os.getcwd()

    rootdir = config.load_swcli_config()["storage"]["root"]
    config_path = config.get_swcli_config_path()
    cmd = [
        "docker",
        "run",
        "--net=host",
        "--rm",
        "-e",
        "DEBUG=1",
        "-e",
        f"SW_USER={gt.getuser()}",
        "-e",
        f"SW_USER_ID={getpwnam(gt.getuser()).pw_uid}",
        "-e",
        "SW_USER_GROUP_ID=0",
        "-e",
        f"SW_LOCAL_STORAGE={rootdir}",
        "-v",
        f"{rootdir}:{rootdir}",
        "-e",
        f"SW_CLI_CONFIG={config_path}",
        "-v",
        f"{config_path}:{config_path}",
        "-v",
        f"{pwd}:{pwd}",
        "-w",
        f"{pwd}",
    ]

    if name:
        cmd += [
            "--name",
            f'"{name}"',
        ]

    if mnt_paths:
        for cp in mnt_paths:
            cmd += [
                "-v",
                f"{cp}:{cp}",
            ]

    if env_vars:
        for _k, _v in env_vars.items():
            cmd.extend(["-e", f"{_k}={_v}"])

    cntr_cache_dir = os.environ.get("SW_PIP_CACHE_DIR", CNTR_DEFAULT_PIP_CACHE_DIR)
    host_cache_dir = os.path.expanduser("~/.cache/starwhale-pip")
    cmd += ["-v", f"{host_cache_dir}:{cntr_cache_dir}"]

    _env = os.environ
    for _ee in (
        "SW_PYPI_INDEX_URL",
        "SW_PYPI_EXTRA_INDEX_URL",
        "SW_PYPI_TRUSTED_HOST",
    ):
        if _ee not in _env:
            continue
        cmd.extend(["-e", f"{_ee}={_env[_ee]}"])

    sw_cmd = " ".join([item for item in sys.argv[1:] if "use-docker" not in item])

    cmd.extend([image, f"swcli {sw_cmd}"])
    return " ".join(cmd)



from __future__ import annotations

import os
import re
import abc
import sys
import copy
import json
import time
import atexit
import base64
import struct
import urllib
import inspect
import zipfile
import binascii
import tempfile
import importlib
import threading
import contextlib
from abc import ABCMeta, abstractmethod
from http import HTTPStatus
from typing import (
    Any,
    Set,
    cast,
    Dict,
    List,
    Type,
    Tuple,
    Union,
    Callable,
    Iterator,
    Optional,
)
from pathlib import Path
from collections import UserDict, OrderedDict

import dill
import numpy as np
import pyarrow as pa  # type: ignore
import requests
import jsonlines
from loguru import logger
from filelock import FileLock
from jsonlines import Writer
from typing_extensions import Protocol

from starwhale.consts import STANDALONE_INSTANCE
from starwhale.utils.fs import ensure_dir
from starwhale.consts.env import SWEnv
from starwhale.utils.dict import flatten as flatten_dict
from starwhale.utils.error import MissingFieldError, FieldTypeOrValueError
from starwhale.utils.retry import http_retry
from starwhale.utils.config import SWCliConfigMixed

datastore_table_file_ext = ".sw-datastore"


class SwType(metaclass=ABCMeta):
    def __init__(self, name: str, pa_type: pa.DataType) -> None:
        self.name = name
        self.pa_type = pa_type

    @abstractmethod
    def merge(self, type: "SwType") -> "SwType":
        ...

    def serialize(self, value: Any) -> Any:
        return value

    def deserialize(self, value: Any) -> Any:
        return value

    @staticmethod
    def encode_schema(type: "SwType") -> Dict[str, Any]:
        if isinstance(type, SwScalarType):
            return {"type": str(type)}
        if isinstance(type, SwListType):
            return {
                "type": "LIST",
                "elementType": SwType.encode_schema(type.element_type),
            }
        if isinstance(type, SwTupleType):
            return {
                "type": "TUPLE",
                "elementType": SwType.encode_schema(type.element_type),
            }
        if isinstance(type, SwMapType):
            return {
                "type": "MAP",
                "keyType": SwType.encode_schema(type.key_type),
                "valueType": SwType.encode_schema(type.value_type),
            }
        if isinstance(type, SwObjectType):
            ret = {
                "type": "OBJECT",
                "attributes": [
                    dict(SwType.encode_schema(v), name=k) for k, v in type.attrs.items()
                ],
            }
            if type.raw_type is Link:
                ret["pythonType"] = "LINK"
            else:
                ret["pythonType"] = (
                    type.raw_type.__module__ + "." + type.raw_type.__name__
                )
            return ret
        raise RuntimeError(f"invalid type {type}")

    @staticmethod
    def decode_schema(schema: Dict[str, Any]) -> "SwType":
        type_name = schema.get("type", None)
        if type_name is None:
            raise RuntimeError("no type in schema")
        if type_name == "LIST":
            element_type = schema.get("elementType", None)
            if element_type is None:
                raise RuntimeError("no element type found for type LIST")
            return SwListType(SwType.decode_schema(element_type))
        if type_name == "TUPLE":
            element_type = schema.get("elementType", None)
            if element_type is None:
                raise RuntimeError("no element type found for type TUPLE")
            return SwTupleType(SwType.decode_schema(element_type))
        if type_name == "MAP":
            key_type = schema.get("keyType", None)
            value_type = schema.get("valueType", None)
            if key_type is None:
                raise RuntimeError("no key type found for type MAP")
            if value_type is None:
                raise RuntimeError("no value type found for type MAP")
            return SwMapType(
                SwType.decode_schema(key_type), SwType.decode_schema(value_type)
            )
        if type_name == "OBJECT":
            raw_type_name = schema.get("pythonType", None)
            if raw_type_name is None:
                raise RuntimeError("no python type found for type OBJECT")
            if raw_type_name == "LINK":
                raw_type = Link
            else:
                parts = raw_type_name.split(".")
                raw_type = getattr(
                    importlib.import_module(".".join(parts[:-1])), parts[-1]
                )
            attrs = {}
            attr_schemas = schema.get("attributes", None)
            if attr_schemas is not None:
                if not isinstance(attr_schemas, list):
                    raise RuntimeError("attributes should be a list")
                for attr in attr_schemas:
                    name: str = attr["name"]
                    if not isinstance(name, str):
                        raise RuntimeError(
                            f"invalid schema, attributes should use strings as names, actual {type(name)}"
                        )
                    attrs[name] = SwType.decode_schema(attr)
            return SwObjectType(raw_type, attrs)
        ret = _TYPE_NAME_DICT.get(type_name, None)
        if ret is None:
            raise RuntimeError(f"can not determine schema: {type_name}")
        return ret

    @abstractmethod
    def encode(self, value: Any) -> Optional[Any]:
        ...

    @abstractmethod
    def decode(self, value: Any) -> Any:
        ...

    @abstractmethod
    def __str__(self) -> str:
        ...

    def __repr__(self) -> str:
        return self.__str__()


class SwScalarType(SwType):
    def __init__(
        self, name: str, pa_type: pa.DataType, nbits: int, default_value: Any
    ) -> None:
        super().__init__(name, pa_type)
        self.nbits = nbits
        self.default_value = default_value

    def merge(self, type: SwType) -> SwType:
        if type is UNKNOWN or self is type:
            return self
        if self is UNKNOWN:
            return type
        raise RuntimeError(f"conflicting type {self} and {type}")

    def encode(self, value: Any) -> Optional[Any]:
        if value is None:
            return None
        if self is UNKNOWN:
            return None
        if self is BOOL:
            if value:
                return "1"
            else:
                return "0"
        if self is STRING:
            return cast(str, value)
        if self is BYTES:
            return base64.b64encode(value).decode()
        if self.name == "int":
            if self.nbits == 8:
                return binascii.hexlify(struct.pack(">b", value)).decode()
            if self.nbits == 16:
                return binascii.hexlify(struct.pack(">h", value)).decode()
            if self.nbits == 32:
                return binascii.hexlify(struct.pack(">i", value)).decode()
            if self.nbits == 64:
                return binascii.hexlify(struct.pack(">q", value)).decode()
        if self.name == "float":
            if self.nbits == 16:
                return binascii.hexlify(struct.pack(">e", value)).decode()
            if self.nbits == 32:
                return binascii.hexlify(struct.pack(">f", value)).decode()
            if self.nbits == 64:
                return binascii.hexlify(struct.pack(">d", value)).decode()
        raise RuntimeError("invalid type " + str(self))

    def decode(self, value: Any) -> Any:
        if value is None:
            return None
        if self is UNKNOWN:
            return None
        if self is BOOL:
            return value == "1"
        if self is STRING:
            return value
        if self is BYTES:
            return base64.b64decode(value)
        if self.name == "int":
            raw = binascii.unhexlify(value.zfill(int(self.nbits / 4)))
            if self.nbits == 8:
                return struct.unpack(">b", raw)[0]
            if self.nbits == 16:
                return struct.unpack(">h", raw)[0]
            if self.nbits == 32:
                return struct.unpack(">i", raw)[0]
            if self.nbits == 64:
                return struct.unpack(">q", raw)[0]
        if self.name == "float":
            raw = binascii.unhexlify(value.zfill(int(self.nbits / 4)))
            if self.nbits == 16:
                return struct.unpack(">e", raw)[0]
            if self.nbits == 32:
                return struct.unpack(">f", raw)[0]
            if self.nbits == 64:
                return struct.unpack(">d", raw)[0]
        raise RuntimeError("invalid type " + str(self))

    def __str__(self) -> str:
        if self.name == "int" or self.name == "float":
            return f"{self.name}{self.nbits}".upper()
        else:
            return self.name.upper()


class SwCompositeType(SwType):
    def __init__(self, name: str) -> None:
        super().__init__(name, pa.binary())

    def serialize(self, value: Any) -> Any:
        return dill.dumps(value)

    def deserialize(self, value: Any) -> Any:
        return dill.loads(value)


class SwListType(SwCompositeType):
    def __init__(self, element_type: SwType) -> None:
        super().__init__("list")
        self.element_type = element_type

    def merge(self, type: SwType) -> SwType:
        if isinstance(type, SwListType):
            t = self.element_type.merge(type.element_type)
            if t is self.element_type:
                return self
            if t is type.element_type:
                return type
            return SwListType(t)
        raise RuntimeError(f"conflicting type {self} and {type}")

    def encode(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, list):
            return [self.element_type.encode(element) for element in value]
        raise RuntimeError(f"value should be a list: {value}")

    def decode(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, list):
            return [self.element_type.decode(element) for element in value]
        raise RuntimeError(f"value should be a list: {value}")

    def __str__(self) -> str:
        return f"[{self.element_type}]"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SwListType):
            return self.element_type == other.element_type
        return False


class SwTupleType(SwCompositeType):
    def __init__(self, element_type: SwType) -> None:
        super().__init__("tuple")
        self.element_type = element_type

    def merge(self, type: SwType) -> SwType:
        if isinstance(type, SwTupleType):
            t = self.element_type.merge(type.element_type)
            if t is self.element_type:
                return self
            if t is type.element_type:
                return type
            return SwTupleType(t)
        raise RuntimeError(f"conflicting type {self} and {type}")

    def encode(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, tuple):
            return tuple([self.element_type.encode(element) for element in value])
        raise RuntimeError(f"value should be a tuple: {value}")

    def decode(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, list):
            return tuple([self.element_type.decode(element) for element in value])
        raise RuntimeError(f"value should be a list: {value}")

    def __str__(self) -> str:
        return f"({self.element_type})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SwTupleType):
            return self.element_type == other.element_type
        return False


class SwMapType(SwCompositeType):
    def __init__(self, key_type: SwType, value_type: SwType) -> None:
        super().__init__("map")
        self.key_type = key_type
        self.value_type = value_type

    def merge(self, type: SwType) -> SwType:
        if isinstance(type, SwMapType):
            kt = self.key_type.merge(type.key_type)
            vt = self.value_type.merge(type.value_type)
            if kt is self.key_type and vt is self.value_type:
                return self
            if kt is type.key_type and vt is type.value_type:
                return type
            return SwMapType(kt, vt)
        raise RuntimeError(f"conflicting type {self} and {type}")

    def encode(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, dict):
            return {
                self.key_type.encode(k): self.value_type.encode(v)
                for k, v in value.items()
            }
        raise RuntimeError(f"value should be a dict: {value}")

    def decode(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, dict):
            return {
                self.key_type.decode(k): self.value_type.decode(v)
                for k, v in value.items()
            }
        raise RuntimeError(f"value should be a dict: {value}")

    def __str__(self) -> str:
        return f"{{{self.key_type}:{self.value_type}}}"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SwMapType):
            return (
                self.key_type == other.key_type and self.value_type == other.value_type
            )
        return False


class SwObjectType(SwCompositeType):
    def __init__(self, raw_type: Type, attrs: Dict[str, SwType]) -> None:
        super().__init__("object")
        self.raw_type = raw_type
        self.attrs = attrs

    def merge(self, type: SwType) -> SwType:
        if type is UNKNOWN or self is type:
            return self
        if isinstance(type, SwObjectType) and self.raw_type is type.raw_type:
            new_attrs: Dict[str, SwType] = {}
            for attr_name, attr_type in self.attrs.items():
                t = type.attrs.get(attr_name, None)
                if t is not None:
                    new_type = attr_type.merge(t)
                    if new_type is not attr_type:
                        new_attrs[attr_name] = new_type
            for attr_name, attr_type in type.attrs.items():
                if attr_name not in self.attrs:
                    new_attrs[attr_name] = attr_type
            if len(new_attrs) == 0:
                return self
            attrs = dict(self.attrs)
            attrs.update(new_attrs)
            return SwObjectType(self.raw_type, attrs)
        raise RuntimeError(f"conflicting type {str(self)} and {str(type)}")

    def encode(self, value: Any) -> Optional[Any]:
        if value is None:
            return None
        if isinstance(value, self.raw_type):
            ret: Dict[str, Any] = {}
            for k, v in value.__dict__.items():
                type = self.attrs.get(k, None)
                if type is None:
                    raise RuntimeError(f"invalid attribute {k}")
                ret[k] = type.encode(v)
            return ret
        raise RuntimeError(
            f"value should be of type {self.raw_type.__name__}, but is {value}"
        )

    def decode(self, value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, dict):
            ret = self.raw_type()
            for k, v in value.items():
                type = self.attrs.get(k, None)
                if type is None:
                    raise RuntimeError(f"invalid attribute {k}")
                ret.__dict__[k] = type.decode(v)
            return ret
        raise RuntimeError(f"value should be a dict: {value}")

    def __str__(self) -> str:
        return (
            self.raw_type.__name__
            + "{"
            + ",".join([f"{k}:{v}" for k, v in self.attrs.items()])
            + "}"
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SwObjectType):
            return self.attrs == other.attrs
        return False


UNKNOWN = SwScalarType("unknown", None, 1, None)
INT8 = SwScalarType("int", pa.int8(), 8, 0)
INT16 = SwScalarType("int", pa.int16(), 16, 0)
INT32 = SwScalarType("int", pa.int32(), 32, 0)
INT64 = SwScalarType("int", pa.int64(), 64, 0)
FLOAT16 = SwScalarType("float", pa.float16(), 16, 0.0)
FLOAT32 = SwScalarType("float", pa.float32(), 32, 0.0)
FLOAT64 = SwScalarType("float", pa.float64(), 64, 0.0)
BOOL = SwScalarType("bool", pa.bool_(), 1, 0)
STRING = SwScalarType("string", pa.string(), 32, "")
BYTES = SwScalarType("bytes", pa.binary(), 32, b"")

_TYPE_DICT: Dict[Any, SwScalarType] = {
    type(None): UNKNOWN,
    np.byte: INT8,
    np.int8: INT8,
    np.int16: INT16,
    np.int32: INT32,
    np.int64: INT64,
    int: INT64,
    np.float16: FLOAT16,
    np.float32: FLOAT32,
    np.float_: FLOAT64,
    float: FLOAT64,
    np.bool_: BOOL,
    bool: BOOL,
    str: STRING,
    bytes: BYTES,
}

_TYPE_NAME_DICT = {str(v): v for v in _TYPE_DICT.values()}


def _get_type(obj: Any) -> SwType:
    element_type: SwType = UNKNOWN
    if isinstance(obj, list):
        for element in obj:
            element_type = element_type.merge(_get_type(element))
        return SwListType(element_type)
    if isinstance(obj, tuple):
        for element in obj:
            element_type = element_type.merge(_get_type(element))
        return SwTupleType(element_type)
    if isinstance(obj, dict):
        key_type: SwType = UNKNOWN
        value_type: SwType = UNKNOWN
        for k, v in obj.items():
            key_type = key_type.merge(_get_type(k))
            value_type = value_type.merge(_get_type(v))
        return SwMapType(key_type, value_type)
    if isinstance(obj, SwObject):
        attrs = {}
        for k, v in obj.__dict__.items():
            attrs[k] = _get_type(v)
        return SwObjectType(type(obj), attrs)
    ret = _TYPE_DICT.get(type(obj), None)
    if ret is None:
        raise RuntimeError(f"unsupported type {type(obj)}")
    return ret


class SwObject:
    def __str__(self) -> str:
        return json.dumps(self._to_dict())

    def __eq__(self, other: Any) -> bool:
        return type(other) is type(self) and self.__dict__ == other.__dict__

    def _to_dict(self) -> Dict[str, Any]:
        ret: Dict[str, Any] = {}
        for k, v in self.__dict__.items():
            if isinstance(v, SwObject):
                ret[k] = v._to_dict()
            elif isinstance(v, bytes):
                ret[k] = BYTES.encode(v)
            else:
                ret[k] = v
        return ret


class Link(SwObject):
    def __init__(
        self,
        uri: Optional[str] = None,
        display_text: Optional[str] = None,
        mime_type: Optional[str] = None,
    ) -> None:
        self.uri = uri
        self.display_text = display_text
        self.mime_type = mime_type


class ColumnSchema:
    def __init__(self, name: str, type: SwType) -> None:
        self.name = name
        self.type = type

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ColumnSchema)
            and self.name == other.name
            and self.type == other.type
        )

    def dumps(self) -> Dict[str, Any]:
        return {"name": self.name, "type": SwType.encode_schema(self.type)}

    @staticmethod
    def loads(obj: Dict[str, Any]) -> ColumnSchema:
        return ColumnSchema(obj["name"], SwType.decode_schema(obj["type"]))


class TableEmptyException(Exception):
    pass


class Record(UserDict):
    def dumps(self) -> Dict[str, Dict]:
        return {
            "schema": {k: SwType.encode_schema(_get_type(v)) for k, v in self.items()},
            "data": {k: _get_type(v).encode(v) for k, v in self.items()},
        }

    @staticmethod
    def loads(obj: Dict[str, Dict]) -> Record:
        schema = obj["schema"]
        data = obj["data"]
        record = Record()
        for k, v in schema.items():
            record[k] = SwType.decode_schema(v).decode(data[k])
        return record


Records = List[Record]


class TableSchemaDesc:
    def __init__(
        self, key_column: Optional[str], columns: Optional[List[ColumnSchema]]
    ) -> None:
        self.key_column = key_column
        self.columns = columns


class TableSchema:
    def __init__(self, key_column: str, columns: List[ColumnSchema]) -> None:
        self.key_column = key_column
        self.columns = {col.name: col for col in columns}

    def copy(self) -> "TableSchema":
        return TableSchema(self.key_column, list(self.columns.values()))

    def merge(self, other: "TableSchema") -> None:
        if self.key_column != other.key_column:
            raise RuntimeError(
                f"conflicting key column, expected {self.key_column}, acutal {other.key_column}"
            )
        new_schema = {}
        for col in other.columns.values():
            column_schema = self.columns.get(col.name, None)
            if column_schema is None:
                new_schema[col.name] = col
            else:
                try:
                    column_schema.type = column_schema.type.merge(col.type)
                except RuntimeError as e:
                    raise RuntimeError(f"can not update column {col.name}") from e
        self.columns.update(new_schema)

    @staticmethod
    def parse(json_str: str) -> "TableSchema":
        d = json.loads(json_str)
        return TableSchema(
            d["key"],
            [
                ColumnSchema(col["name"], SwType.decode_schema(col))
                for col in d["columns"]
            ],
        )

    def __str__(self) -> str:
        columns = []
        for col in self.columns.values():
            d = SwType.encode_schema(col.type)
            d["name"] = col.name
            columns.append(d)
        return json.dumps(
            {
                "key": self.key_column,
                "columns": columns,
            }
        )

    __repr__ = __str__

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, TableSchema)
            and self.key_column == other.key_column
            and self.columns == other.columns
        )


def _get_table_path(root_path: str | Path, table_name: str) -> Path:
    """
    get table path from table name, return the matched file path if there is only one file match the table name
    """
    expect_prefix = Path(root_path) / (table_name.strip("/") + datastore_table_file_ext)
    paths = list(expect_prefix.parent.glob(f"{expect_prefix.name}*"))
    if len(paths) > 1:
        raise RuntimeError(f"can not find table {table_name}, get files {paths}")
    if len(paths) == 1:
        return paths[0]
    return expect_prefix


def _merge_scan(
    iters: List[Iterator[Dict[str, Any]]], keep_none: bool
) -> Iterator[dict]:
    class Node:
        def __init__(self, index: int, iter: Iterator[dict]) -> None:
            self.index = index
            self.iter = iter
            self.item: Optional[Dict[str, Any]] = None
            self.exhausted = False
            self.next_item()

        def next_item(self) -> None:
            try:
                self.item = next(self.iter)
                self.exhausted = False
                self.key = cast(str, self.item["*"])
            except StopIteration:
                self.exhausted = True
                self.item = None
                self.key = ""

    nodes = []
    for i, iter in enumerate(iters):
        node = Node(i, iter)
        if not node.exhausted:
            nodes.append(node)

    while len(nodes) > 0:
        key = min(nodes, key=lambda x: x.key).key
        d: Dict[str, Any] = {}
        for i in range(len(nodes)):
            while nodes[i].key == key:
                item = nodes[i].item
                assert item is not None
                removal = item.pop("-", False)
                item.pop("*", None)
                if removal:
                    d.clear()
                else:
                    d.update(item)
                nodes[i].next_item()
        if len(d) > 0:
            d["*"] = key
            if not keep_none:
                d = {k: v for k, v in d.items() if v is not None}
            yield d
        nodes = [node for node in nodes if not node.exhausted]


def _update_schema(key_column: str, record: Dict[str, Any]) -> TableSchema:
    new_schema = TableSchema(key_column, [])
    for col, value in record.items():
        value_type = _get_type(value)
        new_schema.columns[col] = ColumnSchema(col, value_type)
    return new_schema


class InnerRecord:
    def __init__(self, key_column: str, record: Optional[Record] = None) -> None:
        self.key = ""
        self.key_column = key_column
        self.records: OrderedDict[int, Record] = OrderedDict()
        self.ordered = True
        if record is not None:
            self.append(record)

    def append(self, record: Record) -> str:
        if not self.key:
            self.key = record[self.key_column]
        seq = self._get_seq_num()
        if len(self.records) > 0:
            last = self.get_record()
            # get diff of the last and record
            # let the record only contains the diff or None
            diff = {
                k: v for k, v in record.items() if v is None or last.get(k, None) != v
            }
            if not diff:
                return str(seq)
            record = Record(diff)
        self.ordered = False
        self.records[seq] = record
        return str(seq)

    def update(self, other: InnerRecord | None) -> None:
        if other is None:
            return
        self.records.update(other.records)

    def _reorder(self) -> None:
        if not self.ordered:
            self.records = OrderedDict(sorted(self.records.items()))
            self.ordered = True

    def get_record(
        self, revision: Optional[str] = None, deep_copy: Optional[bool] = None
    ) -> Dict[str, Any]:
        self._reorder()
        ret: Dict[str, Any] = dict()
        for seq, record in self.records.items():
            if revision is None or revision == "" or seq <= int(revision):
                if "-" in record and record["-"]:
                    ret = record.data
                else:
                    if "-" in ret:
                        ret = dict()
                    ret.update(record)
        ret.update({self.key_column: self.key})
        if deep_copy:
            ret = copy.deepcopy(ret)
        return ret

    def dumps(self) -> Dict[str, Any]:
        self._reorder()
        return {
            "key": self.key_column,
            "records": {seq: record.dumps() for seq, record in self.records.items()},
        }

    @staticmethod
    def loads(data: Dict[str, Any]) -> InnerRecord:
        ret = InnerRecord(data["key"])
        ret.ordered = False
        ret.records = OrderedDict(
            {int(seq): Record.loads(record) for seq, record in data["records"].items()}
        )
        first = ret.records[next(iter(ret.records))]

        ret.key = first[data["key"]]
        return ret

    @staticmethod
    def _get_seq_num() -> int:
        return time.monotonic_ns()


class Compressor(abc.ABC):
    @abc.abstractmethod
    def extension(self) -> str:
        """
        Return the extension of the compressed file.
        """
        ...

    @abc.abstractmethod
    def compress(self, source: Path) -> Path:
        """
        Compress the file and return the path to the compressed file.
        """
        ...

    @contextlib.contextmanager
    @abc.abstractmethod
    def decompress(self, source: Path) -> Iterator[Path]:
        """
        Decompress the file and return the path to the temp decompressed file.
        And the temp file will be deleted after the context manager exits.
        """
        ...


class NoCompressor(Compressor):
    def extension(self) -> str:
        return ".json"

    def compress(self, source: Path) -> Path:
        # never be called
        # we need to duplicate the file because the dump method will remove the source file
        raise RuntimeError("should not be called")

    @contextlib.contextmanager
    def decompress(self, source: Path) -> Iterator[Path]:
        # compatible with the existing json file
        yield source


class ZipCompressor(Compressor):
    def extension(self) -> str:
        return ".zip"

    def compress(self, source: Path) -> Path:
        # use the same dir as the source file
        output = source.with_suffix(self.extension())
        with zipfile.ZipFile(output, "w", compression=zipfile.ZIP_DEFLATED) as zipf:
            zipf.write(source, source.name)
        return Path(output)

    @contextlib.contextmanager
    def decompress(self, source: Path) -> Iterator[Path]:
        with zipfile.ZipFile(source, "r") as zipf:
            # extract to tmp dir
            tmp_dir = tempfile.TemporaryDirectory()
            zipf.extractall(tmp_dir.name)
            file_name = zipf.namelist()[0]
            try:
                yield Path(tmp_dir.name) / file_name
            finally:
                tmp_dir.cleanup()


# get all the compressors in this module
compressors: Dict[str, Compressor] = {}
for name, obj in inspect.getmembers(sys.modules[__name__]):
    if inspect.isclass(obj) and issubclass(obj, Compressor) and obj != Compressor:
        compressors[obj.__name__] = obj()


def get_compressor(file: Path) -> Compressor:
    for compressor in compressors.values():
        if file.suffix == compressor.extension():
            return compressor
    raise ValueError(f"Unknown compressor for file {file}")


class MemoryTable:
    def __init__(self, table_name: str, key_column: ColumnSchema) -> None:
        self.table_name = table_name
        self.key_column = key_column
        self.records: Dict[Any, InnerRecord] = {}
        self.lock = threading.Lock()
        self.dirty = False
        self.compressor = ZipCompressor()

    def scan(
        self,
        columns: Optional[Dict[str, str]] = None,
        start: Optional[Any] = None,
        end: Optional[Any] = None,
        keep_none: bool = False,
        end_inclusive: bool = False,
        revision: Optional[str] = None,
        deep_copy: Optional[bool] = None,
    ) -> Iterator[Dict[str, Any]]:
        _end_check: Callable = lambda x, y: x <= y if end_inclusive else x < y
        if deep_copy is None:
            env = os.getenv("SW_DATASTORE_SCAN_DEEP_COPY")
            # make deep copy to True as default if env is not set
            deep_copy = True if env is None else env.strip().upper() == "TRUE"

        with self.lock:
            records = []
            for k, v in self.records.items():
                if (start is None or k >= start) and (
                    end is None or _end_check(k, end)
                ):
                    records.append(v)
        records.sort(key=lambda x: x.key)
        for ir in records:
            r = ir.get_record(revision, deep_copy)
            if columns is None:
                d = dict(r)
            else:
                d = {columns[k]: v for k, v in r.items() if k in columns}
                if "-" in r:
                    d["-"] = r["-"]
            d["*"] = r[self.key_column.name]
            if not keep_none:
                d = {k: v for k, v in d.items() if v is not None}
            yield d

    def insert(self, record: Dict[str, Any]) -> str:
        self.dirty = True
        with self.lock:
            key = record.get(self.key_column.name)
            r = self.records.setdefault(key, InnerRecord(self.key_column.name))
            return r.append(Record(copy.deepcopy(record)))

    def delete(self, keys: List[Any]) -> str | None:
        """
        Delete records by keys. If the key is not found, it will be ignored.
        Returns the sequence number of the last delete operation, or None if no delete operation is performed.
        """
        revision = None
        with self.lock:
            for key in keys:
                r = self.records.get(key, None)
                if r is not None:
                    self.dirty = True
                    revision = r.append(Record({self.key_column.name: key, "-": True}))
        return revision

    def _dump_meta(self) -> Dict[str, Any]:
        return {
            "key_column": self.key_column.dumps(),
            "version": "0.1",
        }

    @classmethod
    def _parse_meta(cls, meta: Any) -> ColumnSchema:
        if not isinstance(meta, dict):
            raise RuntimeError(f"Invalid meta data {meta}")
        if meta["version"] != "0.1":
            raise ValueError(f"Unsupported version {meta['version']}")
        return ColumnSchema.loads(meta["key_column"])

    @classmethod
    def loads(cls, file: Path, table_name: str) -> MemoryTable:
        if not file.exists() or not file.is_file():
            raise RuntimeError(f"File {file} does not exist")

        with get_compressor(file).decompress(file) as f:
            with jsonlines.open(f) as reader:
                meta = reader.read()
                key_column = cls._parse_meta(meta)
                table = MemoryTable(table_name, key_column)
                for record in reader:
                    ir = InnerRecord.loads(record)
                    table.records[ir.key] = ir
        return table

    def dump(self, root_path: str, if_dirty: bool = True) -> None:
        root = Path(root_path)
        lock = root / ".lock"
        with FileLock(lock):
            return self._dump(root, if_dirty)

    def _dump(self, root_path: Path, if_dirty: bool = True) -> None:
        if if_dirty and not self.dirty:
            return

        dst = Path(_get_table_path(root_path, self.table_name))
        base = dst.parent
        temp_filename = base / f"temp.{os.getpid()}"
        ensure_dir(base)

        with jsonlines.open(temp_filename, mode="w") as writer:
            writer.write(self._dump_meta())
            dumped_keys = self._dump_from_local_file(dst, writer)

            for k, ir in self.records.items():
                if k in dumped_keys:
                    continue
                writer.write(ir.dumps())

        compressed = self.compressor.compress(temp_filename)
        os.unlink(temp_filename)
        if dst.suffix == datastore_table_file_ext:
            # the dst file is must not exist, we never save a table as a sw-datastore file
            # use the same extension as compressed file
            ext = datastore_table_file_ext + self.compressor.extension()
        else:
            # the dst file is a compressed file, change the extension
            ext = self.compressor.extension()

        # make dst file have the same extension as compressed file
        new_dst = dst.with_suffix(ext)
        os.rename(compressed, new_dst)
        if new_dst != dst and dst.exists():
            # remove the old file if it is not the new file name
            dst.unlink()
        self.dirty = False

    def _dump_from_local_file(self, existing: Path, output: Writer) -> Set[str]:
        dumped_keys: Set[str] = set()

        if not existing.exists():
            return dumped_keys

        with get_compressor(existing).decompress(existing) as f:
            with jsonlines.open(f, mode="r") as reader:
                self._parse_meta(reader.read())
                for i in reader:
                    ir = InnerRecord.loads(i)
                    r = self.records.get(ir.key)
                    ir.update(r)
                    dumped_keys.add(ir.key)
                    output.write(ir.dumps())

        return dumped_keys


class TableDesc:
    def __init__(
        self,
        table_name: str,
        columns: Union[Dict[str, str], List[str], None] = None,
        keep_none: bool = False,
        revision: Optional[str] = None,
    ) -> None:
        self.table_name = table_name
        self.columns: Optional[Dict[str, str]] = None
        self.keep_none = keep_none
        self.revision = revision
        if columns is not None:
            self.columns = {}
            if isinstance(columns, dict):
                alias_map: Dict[str, str] = {}
                for col, alias in columns.items():
                    key = alias_map.setdefault(alias, col)
                    if key != col:
                        raise RuntimeError(
                            f"duplicate alias {alias} for column {col} and {key}"
                        )
                self.columns = columns
            else:
                for col in columns:
                    if col in self.columns:
                        raise RuntimeError(f"duplicate column name {col}")
                    self.columns[col] = col

    def to_dict(self) -> Dict[str, Any]:
        ret: Dict[str, Any] = {
            "tableName": self.table_name,
        }
        if self.columns is not None:
            ret["columns"] = [
                {"columnName": col, "alias": alias}
                for col, alias in self.columns.items()
            ]
        if self.keep_none:
            ret["keepNone"] = True
        if self.revision is not None:
            ret["revision"] = self.revision
        return ret


class LocalDataStore:
    _instance = None
    _lock = threading.Lock()

    def __str__(self) -> str:
        return f"LocalDataStore, root:{self.root_path}"

    __repr__ = __str__

    @staticmethod
    def get_instance() -> "LocalDataStore":
        with LocalDataStore._lock:
            if LocalDataStore._instance is None:
                ds_path = SWCliConfigMixed().datastore_dir
                ensure_dir(ds_path)
                LocalDataStore._instance = LocalDataStore(str(ds_path))
                atexit.register(LocalDataStore._instance.dump)
            return LocalDataStore._instance

    def __init__(self, root_path: str) -> None:
        self.root_path = root_path
        self.name_pattern = re.compile(r"^[A-Za-z0-9-_/: ]+$")
        self.tables: Dict[str, MemoryTable] = {}
        self.lock = threading.Lock()

    def update_table(
        self,
        table_name: str,
        schema: TableSchema,
        records: List[Dict[str, Any]],
    ) -> str:
        if len(records) == 0:
            return ""
        if self.name_pattern.match(table_name) is None:
            raise RuntimeError(
                f"invalid table name {table_name}, only letters(A-Z, a-z), digits(0-9), hyphen('-'), and underscore('_') are allowed"
            )
        for r in records:
            for k in r.keys():
                if k != "-" and self.name_pattern.match(k) is None:
                    raise RuntimeError(
                        f"invalid column name {k}, only letters(A-Z, a-z), digits(0-9), hyphen('-'), and underscore('_') are allowed"
                    )
        table = self._get_table(table_name, schema.columns[schema.key_column])
        # this will never happen, makes mypy happy
        if table is None:
            raise RuntimeError(
                f"table {table_name} does not exist and can not be created"
            )
        if schema.key_column != table.key_column.name:
            raise RuntimeError(
                f"invalid key column, expected {table.key_column}, actual {schema.key_column}"
            )

        revision = None
        for r in records:
            key = r.get(schema.key_column, None)
            if key is None:
                raise RuntimeError(
                    f"key {schema.key_column} should not be none, record: {r.keys()}"
                )
            if "-" in r:
                revision = table.delete([key])
            else:
                revision = table.insert(r)
        # revision will never be None or empty (len(records) > 0), makes mypy happy
        return revision or ""

    def _get_table(
        self, table_name: str, key_column: ColumnSchema | None, create: bool = True
    ) -> MemoryTable | None:
        with self.lock:
            table = self.tables.get(table_name, None)
            if table is None:
                file = Path(_get_table_path(self.root_path, table_name))
                if file.exists():
                    table = MemoryTable.loads(file, table_name)
                else:
                    if not create:
                        return None
                    if key_column is None:
                        raise RuntimeError(
                            f"key column is required for table {table_name}"
                        )
                    table = MemoryTable(table_name, key_column)
                self.tables[table_name] = table
        return table

    def scan_tables(
        self,
        tables: List[TableDesc],
        start: Optional[Any] = None,
        end: Optional[Any] = None,
        keep_none: bool = False,
        end_inclusive: bool = False,
    ) -> Iterator[Dict[str, Any]]:
        class TableInfo:
            def __init__(
                self,
                name: str,
                key_column_type: SwType,
                columns: Optional[Dict[str, str]],
                keep_none: bool,
                revision: Optional[str],
            ) -> None:
                self.name = name
                self.key_column_type = key_column_type
                self.columns = columns
                self.keep_none = keep_none
                self.revision = revision

        infos: List[TableInfo] = []
        for table_desc in tables:
            table = self._get_table(table_desc.table_name, None, create=False)
            if table is None:
                continue
            key_column_type = table.key_column.type
            infos.append(
                TableInfo(
                    table_desc.table_name,
                    key_column_type,
                    table_desc.columns,
                    table_desc.keep_none,
                    table_desc.revision,
                )
            )

        # check for key type conflicts
        for info in infos:
            if info is infos[0]:
                continue
            if info.key_column_type != infos[0].key_column_type:
                raise RuntimeError(
                    "conflicting key field type. "
                    f"{info.name} has a key of type {info.key_column_type},"
                    f" while {infos[0].name} has a key of type {infos[0].key_column_type}"
                )

        iters = []
        for info in infos:
            iters.append(
                self.tables[info.name].scan(
                    info.columns,
                    start,
                    end,
                    info.keep_none,
                    end_inclusive,
                    revision=info.revision,
                )
            )
        for record in _merge_scan(iters, keep_none):
            record.pop("*", None)
            r: Dict[str, Any] = {}
            for k, v in record.items():
                if k != "*":
                    r[k] = v
            yield r

    def dump(self) -> None:
        for table in list(self.tables.values()):
            table.dump(self.root_path)


class RemoteDataStore:
    def __init__(self, instance_uri: str, token: str) -> None:
        if not instance_uri:
            raise FieldTypeOrValueError("instance_uri not set")

        if not token:
            raise FieldTypeOrValueError("token not set")

        self.instance_uri = instance_uri
        self.token = token

    def __str__(self) -> str:
        return f"RemoteDataStore for {self.instance_uri}"

    __repr__ = __str__

    @http_retry
    def update_table(
        self,
        table_name: str,
        schema: TableSchema,
        records: List[Dict[str, Any]],
    ) -> str:
        data: Dict[str, Any] = {"tableName": table_name}
        column_schemas = []
        for col in schema.columns.values():
            d = SwType.encode_schema(col.type)
            d["name"] = col.name
            column_schemas.append(d)
        data["tableSchemaDesc"] = {
            "keyColumn": schema.key_column,
            "columnSchemaList": column_schemas,
        }
        if records is not None:
            encoded: List[Dict[str, List[Dict[str, Optional[str]]]]] = []
            for record in records:
                r: List[Dict[str, Optional[str]]] = []
                for k, v in record.items():
                    if k == "-":
                        r.append({"key": "-", "value": "1"})
                    else:
                        r.append({"key": k, "value": schema.columns[k].type.encode(v)})
                encoded.append({"values": r})
            data["records"] = encoded

        if self.token is None:
            raise MissingFieldError("no authorization token")

        resp = requests.post(
            urllib.parse.urljoin(self.instance_uri, "/api/v1/datastore/updateTable"),
            data=json.dumps(data, separators=(",", ":")),
            headers={
                "Content-Type": "application/json; charset=utf-8",
                "Authorization": self.token,
            },
            timeout=60,
        )

        if resp.status_code != HTTPStatus.OK:
            logger.error(
                f"[update-table]Table:{table_name}, resp code:{resp.status_code}, \n resp text: {resp.text}, \n records: {records}"
            )
        resp.raise_for_status()
        return resp.json()["data"]  # type: ignore

    @http_retry
    def _do_scan_table_request(self, post_data: Dict[str, Any]) -> Dict[str, Any]:
        resp = requests.post(
            urllib.parse.urljoin(self.instance_uri, "/api/v1/datastore/scanTable"),
            data=json.dumps(post_data, separators=(",", ":")),
            headers={
                "Content-Type": "application/json; charset=utf-8",
                "Authorization": self.token,  # type: ignore
            },
            timeout=60,
        )
        resp.raise_for_status()
        return resp.json()["data"]  # type: ignore

    def scan_tables(
        self,
        tables: List[TableDesc],
        start: Optional[Any] = None,
        end: Optional[Any] = None,
        keep_none: bool = False,
        end_inclusive: bool = False,
    ) -> Iterator[Dict[str, Any]]:
        post_data: Dict[str, Any] = {"tables": [table.to_dict() for table in tables]}
        key_type = _get_type(start)
        if end is not None:
            post_data["end"] = key_type.encode(end)
        if start is not None:
            post_data["start"] = key_type.encode(start)
        post_data["limit"] = 1000
        if keep_none:
            post_data["keepNone"] = True
        if end_inclusive:
            post_data["endInclusive"] = True
        assert self.token is not None
        while True:
            resp_json = self._do_scan_table_request(post_data)
            records = resp_json.get("records", None)
            if records is None or len(records) == 0:
                break
            column_types_list = resp_json.get("columnTypes", None)
            if column_types_list is None:
                raise RuntimeError("no column types in response")
            column_types = {
                col["name"]: SwType.decode_schema(col) for col in column_types_list
            }
            for record in records:
                r: Dict[str, Any] = {}
                for k, v in record.items():
                    col_type = column_types.get(k, None)
                    if col_type is None:
                        raise RuntimeError(
                            f"unknown type for column {k}, record={record}"
                        )
                    r[k] = col_type.decode(v)
                yield r
            if len(records) == 1000:
                post_data["start"] = resp_json["lastKey"]
                post_data["startInclusive"] = False
            else:
                break


class DataStore(Protocol):
    def update_table(
        self,
        table_name: str,
        schema: TableSchema,
        records: List[Dict[str, Any]],
    ) -> str:
        """
        Update a table with records, and return the revision of the table.
        """
        ...

    def scan_tables(
        self,
        tables: List[TableDesc],
        start: Optional[Any] = None,
        end: Optional[Any] = None,
        keep_none: bool = False,
        end_inclusive: bool = False,
    ) -> Iterator[Dict[str, Any]]:
        ...


def get_data_store(instance_uri: str = "", token: str = "") -> DataStore:
    _instance_uri = instance_uri or os.getenv(SWEnv.instance_uri)
    if _instance_uri is None or _instance_uri == STANDALONE_INSTANCE:
        return LocalDataStore.get_instance()
    else:
        token = (
            token
            or SWCliConfigMixed().get_sw_token(instance=instance_uri)
            or os.getenv(SWEnv.instance_token, "")
        )
        return RemoteDataStore(
            instance_uri=_instance_uri,
            token=token,
        )


class TableWriterException(Exception):
    pass


class TableWriter(threading.Thread):
    def __init__(
        self,
        table_name: str,
        key_column: str = "id",
        data_store: Optional[DataStore] = None,
        run_exceptions_limits: int = 100,
    ) -> None:
        super().__init__(name=f"TableWriter-{table_name}")
        self.table_name = table_name
        self.key_column = key_column
        self.data_store = data_store or get_data_store()
        self.latest_revision = ""

        self._cond = threading.Condition()
        self._stopped = False
        self._records: List[Tuple[TableSchema, List[Dict[str, Any]]]] = []
        self._updating_records: List[Tuple[TableSchema, List[Dict[str, Any]]]] = []
        self._queue_run_exceptions: List[Exception] = []
        self._run_exceptions_limits = max(run_exceptions_limits, 0)

        self.daemon = True
        atexit.register(self.close)
        self.start()

    def __enter__(self) -> Any:
        return self

    def __exit__(self, type: Any, value: Any, tb: Any) -> None:
        self.close()

    def close(self) -> None:
        with self._cond:
            if not self._stopped:
                atexit.unregister(self.close)
                self._stopped = True
                self._cond.notify()
        self.join()
        self._raise_run_exceptions(0)

    def _raise_run_exceptions(self, limits: int) -> None:
        if len(self._queue_run_exceptions) > limits:
            _es = self._queue_run_exceptions
            self._queue_run_exceptions = []
            raise TableWriterException(f"{self} run raise {len(_es)} exceptions: {_es}")

    def insert(self, record: Dict[str, Any]) -> None:
        record = flatten_dict(record)
        for k in record:
            for ch in k:
                if (
                    not ch.isalnum()
                    and ch != "-"
                    and ch != "_"
                    and ch != "/"
                    and ch != ":"
                    and not ch.isspace()
                ):
                    raise RuntimeError(f"invalid field {k}")
        self._insert(record)

    def delete(self, key: Any) -> None:
        self._insert({self.key_column: key, "-": True})

    def _insert(self, record: Dict[str, Any]) -> None:
        self._raise_run_exceptions(self._run_exceptions_limits)

        key = record.get(self.key_column, None)
        if key is None:
            raise RuntimeError(
                f"the key {self.key_column} should not be none, record:{record}"
            )
        with self._cond:
            schema = _update_schema(self.key_column, record)
            # TODO: group the records with the same schema
            self._records.append((schema, [record]))
            self._cond.notify()

    def flush(self) -> str:
        """
        Flush the records to the data store, and return the revision of the table.
        """
        while True:
            with self._cond:
                if len(self._records) == 0 and len(self._updating_records) == 0:
                    break
            time.sleep(0.1)
        return self.latest_revision

    def run(self) -> None:
        while True:
            with self._cond:
                while not self._stopped and len(self._records) == 0:
                    self._cond.wait()
                if len(self._records) == 0:
                    break
                self._updating_records = self._records
                self._records = []

            try:
                to_submit: List[Dict[str, Any]] = []
                last_schema = None
                for schema, records in self._updating_records:
                    # group the records with the same schema
                    if last_schema is None:
                        last_schema = schema
                    elif last_schema != schema:
                        self.latest_revision = self.data_store.update_table(
                            self.table_name, last_schema, to_submit
                        )
                        to_submit = []
                        last_schema = schema
                    to_submit.extend(records)
                if len(to_submit) > 0 and last_schema is not None:
                    self.latest_revision = self.data_store.update_table(
                        self.table_name, last_schema, to_submit
                    )
            except Exception as e:
                logger.exception(e)
                self._queue_run_exceptions.append(e)
                if len(self._queue_run_exceptions) > self._run_exceptions_limits:
                    break
            finally:
                self._updating_records = []



import glob
import typing as t
import os.path

from fastapi import APIRouter
from pydantic import Field, BaseModel

from starwhale.api._impl import wrapper
from starwhale.utils.config import SWCliConfigMixed
from starwhale.web.response import success, SuccessResp
from starwhale.api._impl.data_store import SwType, _get_type, TableDesc, LocalDataStore

router = APIRouter()
prefix = "datastore"


class ListTablesRequest(BaseModel):
    prefix: str


class Filter(BaseModel):
    operator: str
    operands: t.List[t.Dict[str, str]]


class QueryTableRequest(BaseModel):
    table_name: str = Field(alias="tableName")
    filter: t.Optional[Filter]
    limit: int


@router.post("/listTables")
def list_tables(request: ListTablesRequest) -> SuccessResp:
    # TODO: use datastore builtin function
    root = str(SWCliConfigMixed().datastore_dir)
    path = os.path.join(root, request.prefix)
    files = glob.glob(f"{path}**", recursive=True)
    files = [
        os.path.split(f)[0][len(root) :].lstrip("/") for f in files if os.path.isfile(f)
    ]
    return success({"tables": files})


@router.post("/queryTable")
def query_table(request: QueryTableRequest) -> SuccessResp:
    eval_id = _is_eval_summary(request)
    if eval_id:
        return _eval_summary(eval_id)

    ds = LocalDataStore.get_instance()
    rows = list(ds.scan_tables([TableDesc(request.table_name)]))
    col, rows = _rows_to_type_and_records(rows)
    return success(
        {
            "columnTypes": col,
            "records": rows,
        }
    )


def _is_eval_summary(request: QueryTableRequest) -> t.Union[str, None]:
    if not request.table_name.endswith("/summary"):
        return None
    if not request.filter:
        return None
    op = request.filter.operator
    operands = request.filter.operands
    if op != "EQUAL":
        return None
    if len(operands) != 2 or {"columnName": "sys/id"} not in operands:
        return None

    eval_id = (
        operands[0].get("intValue")
        if operands[1].get("columnName") == "sys/id"
        else operands[1].get("intValue")
    )

    return eval_id


def _eval_summary(eval_id: str) -> SuccessResp:
    evaluation = wrapper.Evaluation(
        eval_id=eval_id,
        project="self",
        instance="local",
    )
    summary = evaluation.get_metrics()
    col, rows = _rows_to_type_and_records(summary)
    return success(
        {
            "columnTypes": col,
            "records": rows,
        }
    )


def _rows_to_type_and_records(rows: t.Union[list, dict]) -> t.Tuple[list, list]:
    if not rows:
        return [], []
    if not isinstance(rows, list):
        rows = [rows]
    encoders = {k: SwType.encode_schema(_get_type(v)) for k, v in rows[0].items()}
    column_types = [v.update({"name": k}) or v for k, v in encoders.items()]
    return column_types, [{k: str(v) for k, v in row.items()} for row in rows]



from __future__ import annotations

import typing as t
from dataclasses import dataclass

import rich.repr
from rich.text import Text
from textual.widget import Style, Message
from textual.widgets import TreeControl

from starwhale.base.uri import URI
from starwhale.core.instance.view import InstanceTermView
from starwhale.core.project.model import Project

if t.TYPE_CHECKING:
    from textual import events
    from textual._types import MessageTarget
    from textual.widget import RenderableType
    from textual.widgets import NodeID, TreeNode, TreeClick


@dataclass
class ProjectEntry:
    path: str
    is_project: bool


@rich.repr.auto
class ProjectClick(Message, bubble=True):
    def __init__(self, sender: MessageTarget, uri: str) -> None:
        self.uri = uri
        super().__init__(sender)


class ProjectTree(TreeControl[ProjectEntry]):
    def __init__(
        self, uri: str, name: str, current_instance: str = "", current_project: str = ""
    ):
        data = ProjectEntry("", False)
        super().__init__(uri, name=name, data=data)
        self.current_node_id: t.Optional[NodeID] = None
        self.current_instance = current_instance
        self.current_project = current_project

    async def on_mount(self, event: events.Mount) -> None:
        ins = InstanceTermView().list()
        for i in ins:
            await self.root.add(i["name"], ProjectEntry(i["name"], False))
        await self.root.expand()

        # find the node id of current instance
        if self.current_instance:
            cur_ins = None
            for nid in self.root.control.nodes:
                n = self.root.control.nodes[nid]
                if n.data.path == self.current_instance:
                    cur_ins = n
            if cur_ins is not None:
                # expand project list
                await self.add_projects(cur_ins)
                await cur_ins.expand()

                # make current project selected
                if self.current_project:
                    for nid in cur_ins.control.nodes:
                        n = cur_ins.control.nodes[nid]
                        if n.label == self.current_project:
                            await self.emit(ProjectClick(self, n.data.path))
                            self.current_node_id = nid

        self.refresh(layout=True)

    async def handle_tree_click(self, message: TreeClick[ProjectEntry]) -> None:
        dir_entry = message.node.data
        # root entry
        if not dir_entry.path:
            return

        if dir_entry.is_project:
            await self.emit(ProjectClick(self, dir_entry.path))
            self.current_node_id = message.node.id
        else:
            if not message.node.loaded:
                await self.add_projects(message.node)
                await message.node.expand()
                message.node.loaded = True
            else:
                await message.node.toggle()

    async def add_projects(self, node: TreeNode[ProjectEntry]) -> None:
        ins = node.data.path
        ps, _ = Project.list(ins)
        for i in ps:
            proj = URI.project_and_owner_to_uri(i["name"], i.get("owner", ""))
            path = f"{ins}/{proj}"
            await node.add(proj, ProjectEntry(path, True))

    def render_node(self, node: TreeNode[ProjectEntry]) -> RenderableType:
        label = (
            Text(node.label, no_wrap=True, overflow="ellipsis")
            if isinstance(node.label, str)
            else node.label
        )
        if node.id == self.current_node_id:
            label.stylize(Style(bgcolor="magenta"))

        if node.is_cursor:
            label.stylize("reverse")

        meta = {
            "@click": f"click_label({node.id})",
            "tree_node": node.id,
            "cursor": node.is_cursor,
        }
        label.apply_meta(meta)
        return label



from typing import Any, Dict, Tuple, Optional
from dataclasses import dataclass
from urllib.parse import urlparse

from starwhale.utils import config
from starwhale.base.uri import URI
from starwhale.base.uricomponents.exceptions import NoMatchException


def _get_instances() -> Dict[str, Dict]:
    return config.load_swcli_config().get("instances", {})


def _get_default_instance_alias() -> str:
    return config.load_swcli_config().get("current_instance", "")


def _find_alias_by_url(url: str) -> Tuple[str, str]:
    """parse url and return instance alias and path from url"""
    if not url:
        return _get_default_instance_alias(), ""
    p = urlparse(url)

    inst_uri_map = {name: conf["uri"] for name, conf in _get_instances().items()}
    inst_names = list(inst_uri_map.keys())

    # use host as alias when url starts with cloud or non-scheme
    if p.scheme == "cloud":
        if p.netloc not in inst_uri_map:
            raise NoMatchException(p.netloc, inst_names)
        return p.netloc, p.path
    elif p.scheme == "":
        netloc, path = url.split("/", 1)
        if netloc not in inst_uri_map:
            raise NoMatchException(netloc, inst_names)
        return netloc, path
    else:
        ins_url = "://".join([p.scheme, p.netloc])
        hits = [name for name, uri in inst_uri_map.items() if uri == ins_url]
        if len(hits) == 1:
            return hits[0], p.path
        raise NoMatchException(url, hits)


def _check_alias_exists(alias: str) -> None:
    if alias not in _get_instances():
        raise NoMatchException(alias)


@dataclass
class Instance:
    """
    Data structure for Instance info
    """

    alias: str
    path: str = ""

    def __init__(
        self,
        uri: str = "",
        instance_alias: Optional[str] = None,
    ) -> None:
        if instance_alias and uri:
            raise Exception("alias and uri can not both set")
        if not instance_alias:
            instance_alias, path = _find_alias_by_url(uri)
            self.path = path.strip("/")
        _check_alias_exists(instance_alias)
        self.alias = instance_alias

    @property
    def info(self) -> Dict[str, str]:
        """Get current instance info"""
        return _get_instances().get(self.alias, {})

    def __getattr__(self, name: str) -> Any:
        return self.info.get(name)

    @property
    def url(self) -> str:
        return self.info["uri"]

    @property
    def type(self) -> str:
        return self.info["type"]

    @property
    def token(self) -> str:
        return self.info["sw_token"]

    @property
    def is_local(self) -> bool:
        return self.url == "local"

    def __str__(self) -> str:
        if self.is_local:
            return self.url
        return f"cloud://{self.alias}"

    def to_uri(self) -> URI:
        return URI.capsulate_uri(str(self))



import pathlib
from enum import Enum, unique
from dataclasses import dataclass

# TODO: use str path, not Path Class
HOMEDIR = pathlib.Path.home()
CONFIG_DIR = HOMEDIR / ".config" / "starwhale"
SW_CLI_CONFIG = CONFIG_DIR / "config.yaml"

ENV_SW_CLI_CONFIG = "SW_CLI_CONFIG"
ENV_LOG_LEVEL = "SW_LOG_LEVEL"
ENV_LOG_VERBOSE_COUNT = "SW_LOG_VERBOSE_COUNT"
ENV_SW_IMAGE_REPO = "SW_IMAGE_REPO"
# SW_LOCAL_STORAGE env used for generating default swcli config
# and overriding 'storage.root' swcli config in runtime
ENV_SW_LOCAL_STORAGE = "SW_LOCAL_STORAGE"

DEFAULT_STARWHALE_API_VERSION = "1.0"
DEFAULT_MANIFEST_NAME = "_manifest.yaml"

# evaluation related constants
DEFAULT_JOB_NAME = "default"
DEFAULT_EVALUATION_JOB_NAME = "evaluation"
DEFAULT_FINETUNE_JOB_NAME = "fine_tune"
DEFAULT_JOBS_FILE_NAME = "jobs.yaml"
EVALUATION_SVC_META_FILE_NAME = "svc.json"
# auto generated evaluation panel layout file name from yaml or local console
EVALUATION_PANEL_LAYOUT_JSON_FILE_NAME = "eval_panel_layout.json"
# user defined evaluation panel layout file name
EVALUATION_PANEL_LAYOUT_YAML_FILE_NAME = "eval_panel_layout.yaml"

DEFAULT_LOCAL_SW_CONTROLLER_ADDR = "localhost:7827"
LOCAL_CONFIG_VERSION = "2.0"
DEFAULT_FILE_SIZE_THRESHOLD_TO_TAR_IN_MODEL = 10 * 1024  # 10KB

SW_AUTO_DIRNAME = ".starwhale"
SW_EVALUATION_EXAMPLE_DIR = "examples"

# used by the versions before 2.0
# SW_LOCAL_STORAGE = HOMEDIR / ".cache/starwhale"
DEFAULT_SW_LOCAL_STORAGE = HOMEDIR / SW_AUTO_DIRNAME
# SW_TMP_DIR_NAME dir is used for storing the processing files
SW_TMP_DIR_NAME = ".tmp"

ENV_CONDA = "CONDA_DEFAULT_ENV"
ENV_CONDA_PREFIX = "CONDA_PREFIX"
ENV_VENV = "VIRTUAL_ENV"

ENV_POD_NAME = "SW_POD_NAME"
ENV_DISABLE_PROGRESS_BAR = "DISABLE_PROGRESS_BAR"


class DefaultYAMLName:
    MODEL = "model.yaml"
    DATASET = "dataset.yaml"
    RUNTIME = "runtime.yaml"


class PythonRunEnv:
    CONDA = "conda"
    VENV = "venv"
    SYSTEM = "system"
    DOCKER = "docker"
    AUTO = "auto"


class HTTPMethod:
    GET = "GET"
    OPTIONS = "OPTIONS"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


class UserRoleType:
    ADMIN = "admin"
    NORMAL = "normal"


class SupportArch:
    AMD64 = "amd64"
    ARM64 = "arm64"
    NOARCH = "noarch"


class SupportOS:
    UBUNTU = "ubuntu:20.04"


FMT_DATETIME = "%Y-%m-%d %H:%M:%S %Z"

# TODO: use better DEFAULT words?
DEFAULT_COPY_WORKERS = 4

JSON_INDENT = 4

SW_API_VERSION = "v1"

SHORT_VERSION_CNT = 12
VERSION_PREFIX_CNT = 2


class SWDSBackendType:
    S3 = "s3"
    LocalFS = "local_fs"
    SignedUrl = "signed_url"
    Http = "http"


class DataLoaderKind:
    SWDS = "swds"
    JSONL = "jsonl"
    RAW = "raw"


class EvaluationResultKind:
    RESULT = "result"
    METRIC = "metric"


class SWDSSubFileType:
    BIN = "swds_bin"
    META = "swds_meta"


@unique
class FileDesc(Enum):
    MANIFEST = "MANIFEST"
    SRC = "SRC"
    SRC_TAR = "SRC_TAR"
    MODEL = "MODEL"
    DATA = "DATA"


class RunStatus:
    INIT = "init"
    START = "start"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"


class FileFlag:
    UNCHANGED = "unchanged"
    ADDED = "added"
    UPDATED = "updated"
    DELETED = "deleted"


class DecoratorInjectAttr:
    Step = "_starwhale_inject_step_decorator"
    Predict = "_starwhale_inject_predict_decorator"
    Evaluate = "_starwhale_inject_evaluate_decorator"
    FineTune = "_starwhale_inject_finetune_decorator"


@dataclass
class FileNode:
    path: pathlib.Path
    name: str
    size: int
    file_desc: FileDesc
    signature: str = ""
    flag: str = ""


SWDS_DATA_FNAME_FMT = "data_ubyte_{index}.%s" % SWDSSubFileType.BIN
ARCHIVED_SWDS_META_FNAME = "archive.%s" % SWDSSubFileType.META
SWMP_SRC_FNAME = "src.tar"

CURRENT_FNAME = "current"

STANDALONE_INSTANCE = "local"
DEFAULT_INSTANCE = STANDALONE_INSTANCE
DEFAULT_PROJECT = "self"

DEFAULT_PAGE_IDX = 1
DEFAULT_PAGE_SIZE = 20
DEFAULT_REPORT_COLS = 20

RECOVER_DIRNAME = ".recover"
OBJECT_STORE_DIRNAME = ".objectstore"
DATA_STORE_DIRNAME = ".datastore"
DEFAULT_PYTHON_VERSION = "3.8"
LATEST_TAG = "latest"

YAML_TYPES = (".yaml", ".yml")

DEFAULT_IMAGE_REPO = "ghcr.io/star-whale"
SW_IMAGE_FMT = "{repo}/starwhale:{tag}"
DEFAULT_SW_TASK_RUN_IMAGE = SW_IMAGE_FMT.format(repo=DEFAULT_IMAGE_REPO, tag=LATEST_TAG)
SW_IGNORE_FILE_NAME = ".swignore"

CNTR_DEFAULT_PIP_CACHE_DIR = "/root/.cache/pip"

SW_DEV_DUMMY_VERSION = "0.0.0.dev0"
SW_PYPI_PKG_NAME = "starwhale"

DEFAULT_CUDA_VERSION = "11.4.0"
DEFAULT_CONDA_CHANNEL = "conda-forge"

WHEEL_FILE_EXTENSION = ".whl"

CREATED_AT_KEY = "created_at"



from __future__ import annotations

import threading
from typing import Any, Dict, Callable, Optional

from starwhale.consts import DecoratorInjectAttr, DEFAULT_FINETUNE_JOB_NAME


def fine_tune(*args: Any, **kw: Any) -> Any:
    """
    This function can be used as a decorator to define a fine-tune function.

    Argument:
        resources: [Dict, optional] Resources for the predict task, such as memory, gpu etc. Current only supports
            the cloud instance.
    Examples:
    ```python
    from starwhale import experiment

    @experiment.fine_tune
    def ft():
        ...

    ```

    Returns:
        The decorated function
    """

    if len(args) == 1 and len(kw) == 0 and callable(args[0]):
        return fine_tune()(args[0])
    else:

        def _wrap(func: Callable) -> Any:
            _register_ft(func, resources=kw.get("resources"))
            setattr(func, DecoratorInjectAttr.FineTune, True)
            return func

    return _wrap


_registered_ft_func = threading.local()


def _register_ft(
    func: Callable,
    resources: Optional[Dict[str, Any]] = None,
) -> None:
    from .job import step

    try:
        val = _registered_ft_func.value
    except AttributeError:
        val = None

    if val is not None:
        return

    _registered_ft_func.value = step(
        job_name=DEFAULT_FINETUNE_JOB_NAME,
        name="fine_tune",
        resources=resources,
        concurrency=1,
        task_num=1,
    )(func)



from __future__ import annotations

import json
import typing as t
import subprocess
from abc import ABCMeta, abstractmethod
from http import HTTPStatus
from collections import defaultdict

from starwhale.utils import load_yaml
from starwhale.consts import (
    HTTPMethod,
    CREATED_AT_KEY,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
)
from starwhale.base.uri import URI
from starwhale.utils.fs import move_dir, empty_dir
from starwhale.api._impl import wrapper
from starwhale.base.type import InstanceType, JobOperationType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.http import ignore_error
from starwhale.utils.error import NotFoundError, NoSupportError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.utils.process import check_call
from starwhale.core.eval.store import EvaluationStorage
from starwhale.api._impl.metric import MetricKind
from starwhale.core.eval.executor import EvalExecutor
from starwhale.core.runtime.process import Process as RuntimeProcess

_device_id_map = {"cpu": 1, "gpu": 2}


class EvaluationJob(metaclass=ABCMeta):
    def __init__(self, uri: URI) -> None:
        self.uri = uri
        self.name = uri.object.name
        self.project_name = uri.project
        self.sw_config = SWCliConfigMixed()

    @classmethod
    def run(
        cls,
        project_uri: URI,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        version: str = "",
        name: str = "",
        desc: str = "",
        **kw: t.Any,
    ) -> t.Tuple[bool, str]:
        _cls = cls._get_job_cls(project_uri)
        return _cls.run(
            project_uri=project_uri,
            model_uri=model_uri,
            dataset_uris=dataset_uris,
            runtime_uri=runtime_uri,
            version=version,
            name=name,
            desc=desc,
            **kw,
        )

    @abstractmethod
    def info(
        self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE
    ) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    @abstractmethod
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def cancel(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def resume(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def pause(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def compare(self, jobs: t.List[EvaluationJob]) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    @abstractmethod
    def _get_version(self) -> str:
        raise NotImplementedError

    def _get_report(self) -> t.Dict[str, t.Any]:
        evaluation = wrapper.Evaluation(
            eval_id=self._get_version(),
            project=self.uri.project,
            instance=self.uri.instance,
        )
        summary = evaluation.get_metrics()
        kind = summary.get("kind", "")

        ret = {
            "kind": kind,
            "summary": summary,
        }

        if kind == MetricKind.MultiClassification.value:
            ret["labels"] = {
                item["id"]: item for item in list(evaluation.get("labels"))
            }
            ret["confusion_matrix"] = {
                "binarylabel": list(evaluation.get("confusion_matrix/binarylabel"))
            }

        return ret

    @classmethod
    def _get_job_cls(
        cls, uri: URI
    ) -> t.Union[t.Type[StandaloneEvaluationJob], t.Type[CloudEvaluationJob]]:
        if uri.instance_type == InstanceType.STANDALONE:
            return StandaloneEvaluationJob
        elif uri.instance_type == InstanceType.CLOUD:
            return CloudEvaluationJob
        else:
            raise NoSupportError(f"job uri:{uri}")

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _cls = cls._get_job_cls(project_uri)
        return _cls.list(project_uri)

    @classmethod
    def get_job(cls, job_uri: URI) -> EvaluationJob:
        _cls = cls._get_job_cls(job_uri)
        return _cls(job_uri)


# TODO: Storage Class Mixed
class StandaloneEvaluationJob(EvaluationJob):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.store = EvaluationStorage(uri)

    @classmethod
    def run(
        cls,
        project_uri: URI,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        version: str = "",
        name: str = "",
        desc: str = "",
        **kw: t.Any,
    ) -> t.Tuple[bool, str]:
        use_docker = kw.get("use_docker", False)
        step = kw.get("step", "")
        task_index = kw.get("task_index", 0)
        task_num = kw.get("task_num", 0)

        ee = EvalExecutor(
            model_uri=model_uri,
            dataset_uris=dataset_uris,
            project_uri=project_uri,
            runtime_uri=runtime_uri,
            version=version,
            name=name,
            desc=desc,
            step=step,
            task_index=task_index,
            task_num=task_num,
            gencmd=kw.get("gencmd", False),
            use_docker=use_docker,
        )
        if runtime_uri and not use_docker:
            RuntimeProcess.from_runtime_uri(
                uri=runtime_uri,
                target=ee.run,
                args=(),
            ).run()
        else:
            ee.run()

        return True, ee._version

    def _get_version(self) -> str:
        return self.store.id

    @staticmethod
    def _do_flatten_summary(summary: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        rt = {}

        def _f(_s: t.Dict[str, t.Any], _prefix: str = "") -> None:
            for _k, _v in _s.items():
                _k = f"{_prefix}{_k}"
                if isinstance(_v, dict):
                    _f(_v, _prefix=f"{_k}.")
                else:
                    rt[_k] = _v

        _f(summary)
        return rt

    def compare(self, jobs: t.List[EvaluationJob]) -> t.Dict[str, t.Any]:
        rt = {}
        base_report = self._get_report()
        compare_reports = [j._get_report() for j in jobs]

        rt = {
            "kind": base_report["kind"],
            "base": {
                "uri": self.uri.full_uri,
                "version": self.uri.object.name,
            },
            "versions": [self.uri.object.name] + [j.uri.object.name for j in jobs],
            "summary": defaultdict(list),
            "charts": defaultdict(list),
        }

        _base_summary = {}
        _number_types = (int, float, complex)

        for _idx, _report in enumerate([base_report] + compare_reports):
            _summary = _report.get("summary", {})
            if "labels" in _report:
                _summary["labels"] = _report["labels"]

            _flat_summary = self._do_flatten_summary(_summary)
            if _idx == 0:
                _base_summary = _flat_summary

            for _bk, _bv in _base_summary.items():
                _cv = _flat_summary.get(_bk)

                if isinstance(_cv, _number_types) and isinstance(_bv, _number_types):
                    _delta = _cv - _bv
                else:
                    _delta = None
                rt["summary"][_bk].append(
                    {"value": _cv, "delta": _delta, "base": _idx == 0}
                )

            for _k, _v in _report.items():
                if _k in ("kind", "labels", "summary"):
                    continue
                rt["charts"][_k].append(_v)

        return rt

    def info(
        self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE
    ) -> t.Dict[str, t.Any]:

        return {
            "manifest": self.store.manifest,
            "report": self._get_report(),
        }

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        if force:
            empty_dir(self.store.loc)
            return True, ""
        else:
            return move_dir(self.store.loc, self.store.recover_loc, False)

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return move_dir(self.store.recover_loc, self.store.loc, force)

    def cancel(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_container_action(JobOperationType.CANCEL, force)

    def resume(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_container_action(JobOperationType.RESUME, force)

    def pause(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_container_action(JobOperationType.PAUSE, force)

    def _do_container_action(
        self, action: str, force: bool = False
    ) -> t.Tuple[bool, str]:
        cmd = ["docker"]
        if action == JobOperationType.CANCEL:
            cmd += ["rm", "-fv"]
        elif action == JobOperationType.PAUSE:
            cmd += ["pause"]
        elif action == JobOperationType.RESUME:
            cmd += ["unpause"]

        # search container first
        out = subprocess.check_output(
            ["docker", "ps", "-f", f"label=version={self.store.id}", "-q"]
        )
        _container = out.decode().strip()
        if not _container:
            return (
                False,
                f"failed to {action} container, reason: not found container by label=version={self.store.id}",
            )
        cmd += [_container]
        try:
            check_call(cmd)
        except Exception as e:
            return False, f"failed to {action} container, reason: {e}"
        else:
            return True, f"run {action} successfully"

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _rt = []
        for _path, _is_removed in EvaluationStorage.iter_all_jobs(project_uri):
            _manifest = load_yaml(_path)
            if not _manifest:
                continue

            _rt.append(
                {
                    "location": str(_path.absolute()),
                    "manifest": _manifest,
                    "is_removed": _is_removed,
                }
            )
        return _rt, {}


class CloudEvaluationJob(EvaluationJob, CloudRequestMixed):
    @classmethod
    def run(
        cls,
        project_uri: URI,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        version: str = "",
        name: str = "",
        desc: str = "",
        **kw: t.Any,
    ) -> t.Tuple[bool, str]:
        _step_spec = kw.get("step_spec")
        if _step_spec:
            with open(_step_spec) as f:
                _step_spec = f.read()
        crm = CloudRequestMixed()
        # TODO: use argument for uri
        # FIXME: put version into the post fields?
        r = crm.do_http_request(
            f"/project/{project_uri.project}/job",
            method=HTTPMethod.POST,
            instance_uri=project_uri,
            data=json.dumps(
                {
                    "modelVersionUrl": model_uri,
                    "datasetVersionUrls": ",".join([str(i) for i in dataset_uris]),
                    "runtimeVersionUrl": runtime_uri,
                    "stepSpecOverWrites": _step_spec,
                    "resourcePool": kw.get("resource_pool") or "default",
                }
            ),
        )
        if r.status_code == HTTPStatus.OK:
            return True, r.json()["data"]
        else:
            return False, r.json()["message"]

    def _get_version(self) -> str:
        # TODO:use full version id
        return self.uri.object.name

    @ignore_error({})
    def info(
        self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE
    ) -> t.Dict[str, t.Any]:
        if not self.uri.project:
            raise NotFoundError("no selected project")
        return {
            "tasks": self._fetch_tasks(page, size),
            "report": self._get_report(),
            "manifest": self._fetch_job_info(),
        }

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support force remove
        return self.do_http_request_simple_ret(
            f"/project/{self.uri.project}/job/{self.name}",
            method=HTTPMethod.DELETE,
            instance_uri=self.uri,
        )

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.RECOVER, force)

    def cancel(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.CANCEL, force)

    def resume(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.RESUME, force)

    def pause(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.PAUSE, force)

    def _do_job_action(self, action: str, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support force action
        return self.do_http_request_simple_ret(
            f"/project/{self.uri.project}/job/{self.name}/{action}",
            method=HTTPMethod.POST,
            instance_uri=self.uri,
        )

    @classmethod
    @ignore_error(([], {}))
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        if not project_uri.project:
            raise NotFoundError("no selected project")
        crm = CloudRequestMixed()

        r = crm.do_http_request(
            f"/project/{project_uri.project}/job",
            params={"pageNum": page, "pageSize": size},
            instance_uri=project_uri,
        ).json()

        jobs = []
        for j in r["data"]["list"]:
            j.pop("owner", None)
            j[CREATED_AT_KEY] = crm.fmt_timestamp(j["createdTime"])
            j["finished_at"] = crm.fmt_timestamp(j["stopTime"])
            j["duration_str"] = crm.fmt_duration(j["duration"])
            jobs.append({"manifest": j})

        return jobs, crm.parse_pager(r)

    @staticmethod
    def parse_device(device: str) -> t.Tuple[int, float]:
        _t = device.split(":")
        _id = _device_id_map.get(_t[0].lower(), _device_id_map["cpu"])
        _cnt = float(_t[1]) if len(_t) == 2 else 1
        return _id, _cnt

    @ignore_error({})
    def _fetch_job_info(self) -> t.Dict[str, t.Any]:
        r = self.do_http_request(
            f"/project/{self.project_name}/job/{self.name}",
            instance_uri=self.uri,
        ).json()
        return r["data"]  # type: ignore

    @ignore_error(([], {}))
    def _fetch_tasks(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        r = self.do_http_request(
            f"/project/{self.project_name}/job/{self.name}/task",
            params={"pageNum": page, "pageSize": size},
            instance_uri=self.uri,
        ).json()

        tasks = []
        for _t in r["data"]["list"]:
            _t[CREATED_AT_KEY] = self.fmt_timestamp(_t["createdTime"])  # type: ignore
            tasks.append(_t)

        return tasks, self.parse_pager(r)

    def compare(self, jobs: t.List[EvaluationJob]) -> t.Dict[str, t.Any]:
        # TODO: need to implement it
        return {}



import sys
from typing import List

import requests

from .common import (
    random_id,
    FileReader,
    FileWriter,
    TaskRunner,
    ChunkBuffer,
    BrokerReader,
    BrokerWriter,
    PeriodicRunner,
    check_status_code,
)


class CommandRunner(TaskRunner):
    def __init__(self, name: str, broker_base_url: str, args: List[str]):
        super().__init__(name)
        self.broker_base_url = broker_base_url
        self.args = args

    def run(self) -> None:
        stdin_reader = None
        stdin_writer = None
        stdout_reader = None
        stdout_writer = None
        stderr_reader = None
        stderr_writer = None
        keep_alive = None
        try:
            command_id = random_id()
            self.run_until_success(
                lambda: self.new_command(
                    f"{self.broker_base_url}/command", command_id, self.args
                )
            )
            if not self.stopped:
                base_url = f"{self.broker_base_url}/command/{command_id}"
                keep_alive = PeriodicRunner(
                    f"{self.name}-keep-alive", lambda: self.keep_alive(base_url), 1
                )
                keep_alive.start()
                stdin_buffer = ChunkBuffer(256 * 1024, 64)
                stdin_reader = FileReader(
                    f"{self.name}-stdin-reader", sys.stdin, stdin_buffer
                )
                stdin_reader.start()
                stdin_writer = BrokerWriter(
                    f"{self.name}-stdin-writer", f"{base_url}/file/stdin", stdin_buffer
                )
                stdin_writer.start()
                stdout_buffer = ChunkBuffer(256 * 1024, 64)
                stdout_reader = BrokerReader(
                    f"{self.name}-stdout-reader",
                    f"{base_url}/file/stdout",
                    stdout_buffer,
                )
                stdout_reader.start()
                stdout_writer = FileWriter(
                    f"{self.name}-stdout-writer", sys.stdout, stdout_buffer
                )
                stdout_writer.start()
                stderr_buffer = ChunkBuffer(256 * 1024, 64)
                stderr_reader = BrokerReader(
                    f"{self.name}-stderr-reader",
                    f"{base_url}/file/stderr",
                    stderr_buffer,
                )
                stderr_reader.start()
                stderr_writer = FileWriter(
                    f"{self.name}-stderr-writer", sys.stderr, stderr_buffer
                )
                stderr_writer.start()
                while not self.stopped and stdout_writer.is_alive():
                    stdout_writer.join(0.5)
                while not self.stopped and stderr_writer.is_alive():
                    stderr_writer.join(0.5)
        finally:
            if stdin_writer is not None:
                stdin_writer.stop()
                stdin_writer.join()
            if stdout_reader is not None:
                stdout_reader.stop()
                stdout_reader.join()
            if stdout_writer is not None:
                stdout_writer.stop()
                stdout_writer.join()
            if stderr_reader is not None:
                stderr_reader.stop()
                stderr_reader.join()
            if stderr_writer is not None:
                stderr_writer.stop()
                stderr_writer.join()
            if keep_alive is not None:
                keep_alive.stop()
                keep_alive.join()

    def new_command(self, broker_url: str, command_id: str, args: List[str]) -> None:
        resp = requests.post(
            broker_url, json={"command_id": command_id, "args": args}, timeout=5
        )
        check_status_code(resp)

    def keep_alive(self, url: str) -> None:
        requests.post(f"{url}/keepAlive?end=remote", timeout=5)



from __future__ import annotations

import os
import typing as t
import functools
from dataclasses import dataclass

from starwhale.utils import in_production

if t.TYPE_CHECKING:
    from gradio import Blocks
    from gradio.components import Component

Input = t.Union["Component", t.List["Component"]]
Output = t.Union["Component", t.List["Component"]]
Examples = t.Union[t.List[t.Any], str]


@dataclass
class Api:
    input: Input
    output: Output
    func: t.Callable
    uri: str
    examples: t.Optional[Examples]

    def view_func(self, ins: t.Any = None) -> t.Callable:
        func = self.func
        if ins is not None:
            func = functools.partial(func, ins)
        return func


@dataclass
class Hijack:
    """
    Hijack options for online evaluation, useless for local usage.
    """

    # if hijack the submit button js logic
    submit: bool = False
    # the resource path serving on the server side
    # used for example resource render for console
    resource_path: t.Optional[str] = None


class Service:
    def __init__(self, hijack: t.Optional[Hijack] = None) -> None:
        self.apis: t.Dict[str, Api] = {}
        self.api_instance: t.Any = None
        self.example_resources: t.List[str] = []
        self.hijack = hijack

    # TODO: support function as input and output
    def api(
        self,
        _input: Input,
        output: Output,
        uri: t.Optional[str] = None,
        examples: t.Optional[Examples] = None,
    ) -> t.Any:
        def decorator(func: t.Any) -> t.Any:
            self.add_api(_input, output, func, uri or func.__name__, examples)
            return func

        return decorator

    # TODO: support checking duplication
    def add_api(
        self,
        _input: Input,
        output: Output,
        func: t.Callable,
        uri: str,
        examples: t.Optional[Examples] = None,
    ) -> None:
        if not isinstance(_input, list):
            _input = [_input]
        if not isinstance(output, list):
            output = [output]
        _api = Api(_input, output, func, uri, examples)
        self.apis[uri] = _api

    def add_api_instance(self, _api: Api) -> None:
        self.apis[_api.uri] = _api

    def get_spec(self) -> t.Any:
        # fast path
        if not self.apis:
            return {}
        server = self._gen_gradio_server()
        return server.get_config_file()

    def get_openapi_spec(self) -> t.Any:
        server = self._gen_gradio_server()
        return server.app.openapi()

    def _render_api(self, _api: Api) -> None:
        import gradio
        from gradio.components import File, Image, Video, IOComponent

        js_func: t.Optional[str] = None
        if self.hijack and self.hijack.submit:
            js_func = "async(...x) => { typeof wait === 'function' && await wait(); return x; }"
        with gradio.Row():
            with gradio.Column():
                fn = _api.view_func(self.api_instance)
                for i in _api.input:
                    gradio.components.get_component_instance(i, render=False).render()
                submit = gradio.Button("Submit")
                submit.click(fn, inputs=_api.input, outputs=_api.output, _js=js_func)
                # do not serve the useless examples in server instances
                # the console will render them even the models are not serving
                if _api.examples and not in_production():
                    example = gradio.Examples(
                        examples=_api.examples,
                        inputs=[i for i in _api.input if isinstance(i, IOComponent)],
                    )
                    if any(
                        isinstance(i, (File, Image, Video))
                        for i in example.dataset.components
                    ):
                        # examples should be a list of file path
                        # use flatten list
                        to_copy = [i for j in example.examples for i in j]
                        self.example_resources.extend(to_copy)
                        # change example resource path for online evaluation
                        # e.g. /path/to/example.png -> /workdir/src/.starwhale/examples/example.png
                        if self.hijack and self.hijack.resource_path:
                            for i in range(len(example.dataset.samples)):
                                for j in range(len(example.dataset.samples[i])):
                                    origin = example.dataset.samples[i][j]
                                    if origin in to_copy:
                                        name = os.path.basename(origin)
                                        example.dataset.samples[i][j] = os.path.join(
                                            self.hijack.resource_path, name
                                        )
            with gradio.Column():
                for i in _api.output:
                    gradio.components.get_component_instance(i, render=False).render()

    def _gen_gradio_server(self, title: t.Optional[str] = None) -> Blocks:
        import gradio

        apis = self.apis.values()
        with gradio.Blocks() as app:
            with gradio.Tabs():
                for _api in apis:
                    with gradio.TabItem(label=_api.uri):
                        self._render_api(_api)
        app.title = title or "starwhale"
        return app

    def serve(self, addr: str, port: int, title: t.Optional[str] = None) -> None:
        """
        Default serve implementation, users can override this method
        :param addr
        :param port
        :param title webpage title
        :return: None
        """
        server = self._gen_gradio_server(title=title)
        server.launch(server_name=addr, server_port=port)


_svc = Service()


def api(
    _input: Input,
    output: Output,
    uri: t.Optional[str] = None,
    examples: t.Optional[Examples] = None,
) -> t.Any:
    return _svc.api(_input, output, uri, examples)


def internal_api_list() -> t.Dict[str, Api]:
    return _svc.apis



import os
import sys
import time
import typing as t
import platform
import threading
from pathlib import Path
from collections import defaultdict

import psutil
from loguru import logger

from starwhale.consts import FMT_DATETIME
from starwhale.utils.dict import flatten
from starwhale.utils.venv import guess_current_py_env
from starwhale.utils.error import NotFoundError

from .base import _TrackSource
from .tracker import Tracker


class CollectorThread(threading.Thread):
    def __init__(
        self,
        tracker: Tracker,
        workdir: t.Union[Path, str],
        sample_interval: float = 1.0,
        report_interval: float = 30.0,
        run_exceptions_limits: int = 100,
    ) -> None:
        super().__init__(name="CollectorThread")
        self.tracker = tracker
        workdir = Path(workdir)
        if not workdir.exists():
            raise NotFoundError(f"Collector workdir({workdir}) not found")
        self.workdir = workdir

        if sample_interval > report_interval:
            raise ValueError(
                f"sample interval({sample_interval}) should be less than report interval({report_interval})"
            )
        if sample_interval <= 0 or sample_interval > 60:
            raise ValueError(
                f"sample interval seconds({sample_interval}) must be between 0 and 60"
            )
        self.sample_interval = sample_interval

        if report_interval <= 0 or report_interval > 300:
            raise ValueError(
                f"report interval seconds({report_interval}) must be between 0 and 300"
            )
        self.report_interval = report_interval

        self._pid = os.getpid()
        self._staging_metrics: t.Dict = defaultdict(list)
        self._metrics_step = 0
        self._metrics_inspect_cnt = 0
        self._stopped = True
        self._stop_event = threading.Event()
        self._run_exceptions: t.List[Exception] = []

        self._run_exceptions_limits: int = max(run_exceptions_limits, 0)
        self.daemon = True

    def _raise_run_exceptions(self, limits: t.Optional[int] = None) -> None:
        limits = self._run_exceptions_limits if limits is None else limits
        if len(self._run_exceptions) > limits:
            raise threading.ThreadError(
                f"{self} run raise {len(self._run_exceptions)} exceptions: {self._run_exceptions}"
            )

    def run(self) -> None:
        self._stopped = False
        self._stop_event.clear()

        inspect_cases = [
            ("code info", self._inspect_code),
            ("python run info", self._inspect_python_run),
            ("system specs", self._inspect_system_specs),
            ("process info", self._inspect_process_specs),
            ("environments", self._inspect_environments),
        ]
        for _info, _action in inspect_cases:
            try:
                ret = _action()
                self.tracker._log_params(ret, source=_TrackSource.SYSTEM)
            except Exception as e:
                logger.exception(f"failed to inspect {_info}: {e}")

        # TODO: tune the accuracy of inspect and report interval
        last_inspect_time = last_report_time = time.monotonic()
        check_interval = min(0.1, self.sample_interval)
        while True:
            if self._stop_event.wait(timeout=check_interval):
                break

            try:
                if time.monotonic() - last_inspect_time > self.sample_interval:
                    metrics = self._inspect_metrics()
                    for k, v in flatten(metrics).items():
                        self._staging_metrics[k].append(float(v))
                    last_inspect_time = time.monotonic()
                    self._metrics_inspect_cnt += 1

                if time.monotonic() - last_report_time > self.report_interval:
                    self._report_metrics()
                    last_report_time = time.monotonic()
            except Exception as e:
                logger.exception(e)
                self._run_exceptions.append(e)
                self._raise_run_exceptions()

    def close(self) -> None:
        self._stop_event.set()
        self.join()
        self._stopped = True
        self._staging_metrics = defaultdict(list)
        self._raise_run_exceptions(0)

    def _inspect_code(self) -> t.Dict[str, t.Any]:
        # TODO: log code file as artifacts
        # TODO: support users custom to upload code scripts
        # TODO: make git fetch robust
        import git

        def _get_git_info() -> t.Dict[str, t.Any]:
            repo = git.Repo(self.workdir, search_parent_directories=True)
            commit = repo.head.commit
            return {
                "commit_sha": commit.hexsha,
                "message": commit.message,
                "author_name": commit.author.name,
                "author_email": commit.author.email,
                "date": commit.committed_datetime.strftime(FMT_DATETIME),
                "is_dirty": repo.is_dirty(untracked_files=True),
                "branch": repo.active_branch.name,
                "remote_urls": [r.url for r in repo.remotes],
            }

        try:
            git_info = _get_git_info()
        except git.InvalidGitRepositoryError:
            git_info = {}

        return {
            "code": {
                "git": git_info,
                "workdir": str(self.workdir),
            }
        }

    def _inspect_python_run(self) -> t.Dict[str, t.Any]:
        # TODO: add starwhale runtime, pip dependencies(pip requirements.txt and conda export.yaml) info
        return {
            "python": {
                "version": platform.python_version(),
                "bin_path": sys.executable,
                "env_mode": guess_current_py_env(),
            }
        }

    def _inspect_system_specs(self) -> t.Dict[str, t.Any]:
        # TODO: add more system specs: gpu, network, ip...
        return {
            "system": {
                "hostname": platform.node(),
                "platform": sys.platform,
                "os": platform.platform(),
                "arch": platform.machine(),
            },
            "hardware": {
                "cpu": {
                    "cores": psutil.cpu_count(logical=False),
                    "cores_logical": psutil.cpu_count(logical=True),
                },
                "memory": {"total_bytes": psutil.virtual_memory().total},
                "disk": {
                    "total_bytes": psutil.disk_usage("/").total,
                    "used_bytes": psutil.disk_usage("/").used,
                },
            },
        }

    def _inspect_process_specs(self) -> t.Dict[str, t.Any]:
        # TODO: get process more info
        return {
            "process": {
                "pid": self._pid,
                "cwd": str(self.workdir),
                "cmdline": " ".join(psutil.Process(self._pid).cmdline()),
            },
        }

    def _inspect_environments(self) -> t.Dict[str, t.Any]:
        return {
            "environments": dict(os.environ),
        }

    def _inspect_metrics(self) -> t.Dict[str, t.Any]:
        _process = psutil.Process(self._pid)
        # TODO: support more system and process metrics: network, io, memory
        system_cpu_percent = [float(v) for v in psutil.cpu_percent(percpu=True)]  # type: ignore
        cpu_cores = float(len(system_cpu_percent))
        return {
            "system": {
                "cpu": {
                    "usage_percent": sum(system_cpu_percent) / cpu_cores,
                },
            },
            "process": {
                "cpu": {
                    "usage_percent": _process.cpu_percent() / cpu_cores,
                },
                "num_threads": _process.num_threads(),
            },
        }

    def _report_metrics(self) -> None:
        data = {}
        for k, v in self._staging_metrics.items():
            if not v:
                continue

            data[f"{k}/last"] = float(v[-1])
            data[f"{k}/avg"] = round(sum(v) / len(v), 3)

        if not data:
            return

        self.tracker._log_metrics(
            data=data,
            step=self._metrics_step,
            commit=True,
            source=_TrackSource.SYSTEM,
        )

        self._metrics_step += 1
        self._staging_metrics = defaultdict(list)
